import struct

from bxgateway.btc_constants import BTC_BLOCK_HDR_SIZE, BTC_SHA_HASH_LEN, BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import pack_int_to_btc_varint
from bxcommon.utils import crypto
from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int


# A BlockHeader is the first 80 bytes of the corresponding block message payload
# terminated by a null byte (\x00) to signify that there are no services.
# FIXME, there's a lot of duplicate code between here and BlockBTCMessage
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash


class BlockHeader(object):
    def __init__(self, buf=None, version=None, prev_block=None, merkle_root=None,
                 timestamp=None, bits=None, nonce=None):
        if buf is None:
            buf = bytearray(81)
            self.buf = buf
            off = 0
            struct.pack_into('<I', buf, off, version)
            off += 4
            buf[off:off + 32] = prev_block.get_little_endian()
            off += 32
            buf[off:off + 32] = merkle_root.get_little_endian()
            off += 32
            struct.pack_into('<III', buf, off, timestamp, bits, nonce)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)

        self._version = self._prev_block = self._merkle_root = None
        self._timestamp = self._bits = self._nonce = None
        self._block_hash = None
        self._txn_count = None
        self._hash_val = None

    def version(self):
        if self._version is None:
            off = 0
            self._version = struct.unpack_from('<I', self.buf, off)[0]
            off += 4
            self._prev_block = BtcObjectHash(self.buf, off, 32)
            off += 32
            self._merkle_root = self._memoryview[off:off + 32]
            off += 32
            self._timestamp, self._bits, self._nonce = struct.unpack_from('<III', self.buf, off)
            self._txn_count, size = btc_varint_to_int(self.buf, off)
        return self._version

    def prev_block_hash(self):
        if self._version is None:
            self.version()
        return self._prev_block

    def merkle_root(self):
        if self._version is None:
            self.version()
        return self._merkle_root

    def timestamp(self):
        if self._version is None:
            self.version()
        return self._timestamp

    def bits(self):
        if self._version is None:
            self.version()
        return self._bits

    def nonce(self):
        if self._version is None:
            self.version()
        return self._nonce

    def block_hash(self):
        if self._block_hash is None:
            header = self._memoryview[:BTC_BLOCK_HDR_SIZE]  # remove the tx count at the end
            raw_hash = crypto.bitcoin_hash(header)
            self._hash_val = BtcObjectHash(buf=raw_hash, length=BTC_SHA_HASH_LEN)
        return self._hash_val


class HeadersBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.HEADERS

    def __init__(self, magic=None, headers=None, buf=None):
        if buf is None:
            buf = bytearray(BTC_HDR_COMMON_OFF + 9 + len(headers) * 81)
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            off += pack_int_to_btc_varint(len(headers), buf, off)
            for header in headers:
                buf[off:off + 81] = header
                off += 81

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._headers = self._header_count = self._block_hash = None

    def hash_count(self):
        if self._header_count is None:
            off = BTC_HDR_COMMON_OFF
            self._header_count, size = btc_varint_to_int(self.buf, off)

        return self._header_count

    def __iter__(self):
        off = BTC_HDR_COMMON_OFF
        self._header_count, size = btc_varint_to_int(self.buf, off)
        off += size
        for _ in range(self._header_count):
            yield self._memoryview[off:off + 81]
            off += 81



import blxr_rlp as rlp
from bxgateway.messages.eth.discovery.eth_discovery_message import EthDiscoveryMessage
from bxgateway.messages.eth.discovery.eth_discovery_message_type import EthDiscoveryMessageType
from bxgateway.messages.eth.serializers.endpoint import Endpoint


class PingEthDiscoveryMessage(EthDiscoveryMessage):
    msg_type = EthDiscoveryMessageType.PING

    fields = [("protocol_version", rlp.sedes.big_endian_int),
              ("listen_address", Endpoint()),
              ("remote_address", Endpoint()),
              ("expiration", rlp.sedes.big_endian_int)]

    def get_protocol_version(self):
        return self.get_field_value("protocol_version")

    def get_listen_address(self):
        return self.get_field_value("listen_address")

    def get_remote_address(self):
        return self.get_field_value("remote_address")

    def get_expiration(self):
        return self.get_field_value("expiration")



from dataclasses import dataclass
from typing import Type, Dict, Any, Optional, TYPE_CHECKING

from prometheus_client import Counter

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.bloxroute.bdn_performance_stats_message import BdnPerformanceStatsData
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.utils.stats.statistics_service import StatisticsService, StatsIntervalData
from bxgateway import gateway_constants
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


@dataclass
class GatewayBdnPerformanceStatInterval(StatsIntervalData):
    blockchain_node_to_bdn_stats: Optional[Dict[IpEndpoint, BdnPerformanceStatsData]] = None

    def __init__(self):
        super().__init__()
        self.blockchain_node_to_bdn_stats = {}


blocks_from_bdn = Counter(
    "blocks_from_bdn", "Number of blocks received first from the BDN", ['ip_endpoint']
)
blocks_from_blockchain = Counter(
    "blocks_from_blockchain", "Number of blocks received first from the blockchain node", ['ip_endpoint']
)
blocks_seen = Counter(
    "blocks_seen", "Total number of unique new block seen by gateway", ['ip_endpoint']
)
transactions_from_bdn = Counter(
    "transactions_from_bdn", "Number of transactions received first from the BDN", ['ip_endpoint']
)
transactions_from_blockchain = Counter(
    "transactions_from_blockchain", "Number of transactions received first from the blockchain node", ['ip_endpoint']
)


class _GatewayBdnPerformanceStatsService(
    StatisticsService[GatewayBdnPerformanceStatInterval, "AbstractGatewayNode"]
):
    def __init__(
        self,
        interval=gateway_constants.GATEWAY_BDN_PERFORMANCE_STATS_INTERVAL_S,
        look_back=gateway_constants.GATEWAY_BDN_PERFORMANCE_STATS_LOOKBACK,
    ) -> None:
        super(_GatewayBdnPerformanceStatsService, self).__init__(
            "GatewayBdnPerformanceStats",
            interval,
            look_back,
            reset=True,
            stat_logger=logging.get_logger(LogRecordType.BdnPerformanceStats, __name__),
        )

    def create_interval_data_object(self) -> None:
        super().create_interval_data_object()

        node = self.node
        assert node is not None
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None

        active_blockchain_peers = list(node.connection_pool.get_by_connection_types((ConnectionType.BLOCKCHAIN_NODE,)))
        for blockchain_peer in active_blockchain_peers:
            blockchain_node_to_bdn_stats[blockchain_peer.endpoint] = BdnPerformanceStatsData()

    def get_node_stats(
        self, node_endpoint: IpEndpoint, blockchain_node_to_bdn_stats: Dict[IpEndpoint, BdnPerformanceStatsData]
    ) -> BdnPerformanceStatsData:
        if node_endpoint in blockchain_node_to_bdn_stats:
            node_stats = blockchain_node_to_bdn_stats[node_endpoint]
        else:
            blockchain_node_to_bdn_stats[node_endpoint] = BdnPerformanceStatsData()
            node_stats = blockchain_node_to_bdn_stats[node_endpoint]
        return node_stats

    def get_interval_data_class(self) -> Type[GatewayBdnPerformanceStatInterval]:
        return GatewayBdnPerformanceStatInterval

    def get_info(self) -> Dict[str, Any]:
        return {}

    def log_block_from_blockchain_node(self, node_endpoint: IpEndpoint) -> None:
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None
        node_stats = self.get_node_stats(node_endpoint, blockchain_node_to_bdn_stats)

        node_stats.new_blocks_received_from_blockchain_node += 1
        node_stats.new_blocks_seen += 1
        blocks_from_blockchain.labels(f"{node_endpoint.ip_address}:{node_endpoint.port}").inc()
        blocks_seen.labels(f"{node_endpoint.ip_address}:{node_endpoint.port}").inc()

        for endpoint, stats in blockchain_node_to_bdn_stats.items():
            if endpoint == node_endpoint:
                continue
            stats.new_blocks_received_from_bdn += 1
            stats.new_blocks_seen += 1
            blocks_from_bdn.labels(f"{endpoint.ip_address}:{endpoint.port}").inc()
            blocks_seen.labels(f"{endpoint.ip_address}:{endpoint.port}").inc()

    def log_block_message_from_blockchain_node(self, node_endpoint: IpEndpoint, is_full_block: bool) -> None:
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None
        node_stats = self.get_node_stats(node_endpoint, blockchain_node_to_bdn_stats)

        if is_full_block:
            node_stats.new_block_messages_from_blockchain_node += 1
        else:
            node_stats.new_block_announcements_from_blockchain_node += 1

    def log_block_from_bdn(self) -> None:
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None

        for endpoint, stats in blockchain_node_to_bdn_stats.items():
            stats.new_blocks_received_from_bdn += 1
            stats.new_blocks_seen += 1
            blocks_from_bdn.labels(f"{endpoint.ip_address}:{endpoint.port}").inc()
            blocks_seen.labels(f"{endpoint.ip_address}:{endpoint.port}").inc()

    def log_tx_from_blockchain_node(self, node_endpoint: IpEndpoint, low_fee: bool = False) -> None:
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None
        node_stats = self.get_node_stats(node_endpoint, blockchain_node_to_bdn_stats)

        transactions_from_blockchain.labels(f"{node_endpoint.ip_address}:{node_endpoint.port}").inc()

        if low_fee:
            node_stats.new_tx_received_from_blockchain_node_low_fee += 1
        else:
            node_stats.new_tx_received_from_blockchain_node += 1
            for endpoint, stats in blockchain_node_to_bdn_stats.items():
                if endpoint == node_endpoint:
                    continue
                stats.new_tx_received_from_bdn += 1
                transactions_from_bdn.labels(f"{endpoint.ip_address}:{endpoint.port}").inc()

    def log_tx_from_bdn(self, low_fee: bool = False) -> None:
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None

        for endpoint, stats in blockchain_node_to_bdn_stats.items():
            transactions_from_bdn.labels(f"{endpoint.ip_address}:{endpoint.port}").inc()
            if low_fee:
                stats.new_tx_received_from_bdn_low_fee += 1
            else:
                stats.new_tx_received_from_bdn += 1

    def log_tx_sent_to_nodes(self, broadcasting_endpoint: Optional[IpEndpoint] = None):
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None

        for endpoint, stats in blockchain_node_to_bdn_stats.items():
            if broadcasting_endpoint is not None and endpoint == broadcasting_endpoint:
                continue
            stats.tx_sent_to_node += 1

    def log_duplicate_tx_from_node(self, node_endpoint: IpEndpoint):
        blockchain_node_to_bdn_stats = self.interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None
        node_stats = self.get_node_stats(node_endpoint, blockchain_node_to_bdn_stats)
        node_stats.duplicate_tx_from_node += 1

    def get_most_recent_stats(self) -> Optional[GatewayBdnPerformanceStatInterval]:
        if self.history:
            interval_data = self.history[0]
            return interval_data
        else:
            return None


gateway_bdn_performance_stats_service = _GatewayBdnPerformanceStatsService()



from typing import Optional, Union, List, NamedTuple, Tuple, Dict
from abc import abstractmethod
from datetime import datetime
import time

from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.services.transaction_service import TransactionService
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.utils import crypto, convert
from bxcommon.utils.blockchain_utils.bdn_tx_to_bx_tx import bdn_tx_to_bx_tx
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.proxy.vector_proxy import VectorProxy
from bxcommon import constants as common_constants

from bxgateway import btc_constants
from bxgateway.abstract_message_converter import AbstractMessageConverter, BlockDecompressionResult
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.compact_block_btc_message import CompactBlockBtcMessage
from bxgateway.messages.btc.tx_btc_message import TxBtcMessage
from bxgateway.utils.block_info import BlockInfo


class CompactBlockCompressionResult:
    def __init__(
        self,
        success: bool,
        block_info: Optional[BlockInfo],
        bx_block: Optional[Union[memoryview, bytearray]],
        recovery_index: Optional[int],
        missing_indices: List[int],
        recovered_transactions: Union[List[memoryview], VectorProxy]
     ):
        self.success = success
        self.block_info = block_info
        self.bx_block = bx_block
        self.recovery_index = recovery_index
        self.missing_indices = missing_indices
        self.recovered_transactions = recovered_transactions


class CompactBlockRecoveryData(NamedTuple):
    block_transactions: List[Optional[Union[memoryview, int]]]
    block_header: memoryview
    magic: int
    tx_service: TransactionService


def get_block_info(
        bx_block: memoryview,
        block_hash: Sha256Hash,
        short_ids: List[int],
        decompress_start_datetime: datetime,
        decompress_start_timestamp: float,
        total_tx_count: Optional[int] = None,
        btc_block_msg: Optional[BlockBtcMessage] = None
) -> BlockInfo:
    if btc_block_msg is not None:
        bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(bx_block))
        compressed_size = len(bx_block)
        prev_block_hash = convert.bytes_to_hex(btc_block_msg.prev_block_hash().binary)
        btc_block_len = len(btc_block_msg.rawbytes())
        compression_rate = 100 - float(compressed_size) / btc_block_len * 100
    else:
        bx_block_hash = None
        compressed_size = None
        prev_block_hash = None
        btc_block_len = None
        compression_rate = None
    return BlockInfo(
        block_hash,
        short_ids,
        decompress_start_datetime,
        datetime.utcnow(),
        (time.time() - decompress_start_timestamp) * 1000,
        total_tx_count,
        bx_block_hash,
        prev_block_hash,
        btc_block_len,
        compressed_size,
        compression_rate,
        []
    )


class AbstractBtcMessageConverter(AbstractMessageConverter):

    def __init__(self, btc_magic):
        if not btc_magic:
            raise ValueError("btc_magic is required")

        self._btc_magic = btc_magic
        self._last_recovery_idx: int = 0
        self._recovery_items: Dict[int, CompactBlockRecoveryData] = {}

    @abstractmethod
    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Uncompresses a bx_block from a broadcast bx_block message and converts to a raw BTC bx_block.

        bx_block must be a memoryview, since memoryview[offset] returns a bytearray, while bytearray[offset] returns
        a byte.
        """
        pass

    @abstractmethod
    def compact_block_to_bx_block(
            self,
            compact_block: CompactBlockBtcMessage,
            transaction_service: TransactionService
    ) -> CompactBlockCompressionResult:
        """
         Handle decompression of Bitcoin compact block.
         Decompression converts compact block message to full block message.
         """
        pass

    @abstractmethod
    def recovered_compact_block_to_bx_block(
            self,
            failed_compression_result: CompactBlockCompressionResult,
    ) -> CompactBlockCompressionResult:
        pass

    def bx_tx_to_tx(self, tx_msg):
        if not isinstance(tx_msg, TxMessage):
            raise TypeError("tx_msg is expected to be of type TxMessage")

        buf = bytearray(btc_constants.BTC_HDR_COMMON_OFF) + tx_msg.tx_val()
        raw_btc_tx_msg = BtcMessage(self._btc_magic, TxBtcMessage.MESSAGE_TYPE, len(tx_msg.tx_val()), buf)
        btc_tx_msg = TxBtcMessage(buf=raw_btc_tx_msg.buf)

        return btc_tx_msg

    def tx_to_bx_txs(
        self,
        tx_msg,
        network_num: int,
        transaction_flag: Optional[TransactionFlag] = None,
        min_tx_network_fee: int = 0,
        account_id: str = common_constants.DECODED_EMPTY_ACCOUNT_ID
    ) -> List[Tuple[TxMessage, Sha256Hash, Union[bytearray, memoryview]]]:
        if not isinstance(tx_msg, TxBtcMessage):
            raise TypeError("tx_msg is expected to be of type TxBTCMessage")

        bx_tx_msg = TxMessage(
            tx_msg.tx_hash(),
            network_num,
            tx_val=tx_msg.tx(),
            transaction_flag=transaction_flag,
            account_id=account_id
        )

        return [(bx_tx_msg, tx_msg.tx_hash(), tx_msg.tx())]

    def bdn_tx_to_bx_tx(
        self,
        raw_tx: Union[bytes, bytearray, memoryview],
        network_num: int,
        transaction_flag: Optional[TransactionFlag] = None,
        account_id: str = common_constants.DECODED_EMPTY_ACCOUNT_ID
    ) -> TxMessage:
        return bdn_tx_to_bx_tx(raw_tx, network_num, transaction_flag, account_id)



import typing
from abc import abstractmethod
from typing import Iterable, TYPE_CHECKING

from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway.messages.eth.protocol.new_block_eth_protocol_message import NewBlockEthProtocolMessage
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.eth.eth_block_queuing_service import EthBlockQueuingService

from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(LogRecordType.BlockCleanup, __name__)


class AbstractEthBlockCleanupService(AbstractBlockCleanupService):
    """
    Service for managing block cleanup.
    """

    def __init__(self, node: "EthGatewayNode", network_num: int) -> None:
        """
        Constructor
        :param node: reference to node object
        :param network_num: network number
        """

        super(AbstractEthBlockCleanupService, self).__init__(node=node, network_num=network_num)


    def clean_block_transactions(
        self,
        block_msg: NewBlockEthProtocolMessage,
        transaction_service: TransactionService
    ) -> None:
        block_hash = block_msg.block_hash()
        transactions_list = block_msg.txns()
        self.clean_block_transactions_by_block_components(
            block_hash=block_hash,
            transactions_list=(tx.hash() for tx in transactions_list),
            transaction_service=transaction_service
        )

    def clean_block_transactions_from_block_queue(
        self,
        block_hash: Sha256Hash,
        block_queuing_service: EthBlockQueuingService
    ) -> None:
        try:
            block_body = block_queuing_service.get_block_body_from_message(block_hash)
            assert block_body is not None
            transactions_hashes = block_body.get_block_transaction_hashes(0)
            self.clean_block_transactions_by_block_components(
                transaction_service=self.node.get_tx_service(),
                block_hash=block_hash,
                transactions_list=transactions_hashes
            )
        except Exception as error:
            logger.error("clean_block_transactions_from_block_queue operation for block {} failed with error {}.",
                         block_hash, error)
            block_body_parts = block_queuing_service.get_block_parts(block_hash)
            logger.debug("Failed block body bytes: {}",
                         "Empty" if block_body_parts is None
                         else convert.bytes_to_hex(block_body_parts.block_body_bytes))

    @abstractmethod
    def clean_block_transactions_by_block_components(
        self,
        block_hash: Sha256Hash,
        transactions_list: Iterable[Sha256Hash],
        transaction_service: TransactionService
    ) -> None:
        pass

    def _request_block(self, block_hash: Sha256Hash) -> None:
        node_conn = typing.cast("bxgateway.connections.eth.eth_node_connection.EthNodeConnection",
                                self.node.get_any_active_blockchain_connection())
        if node_conn is None:
            logger.debug("Request for block '{}' failed. No connection to node.", repr(block_hash))
            return
        connection_protocol = node_conn.connection_protocol
        connection_protocol.request_block_body([block_hash])
        logger.trace("Block cleanup request for {}", block_hash)



import struct
from typing import Tuple, Optional, Union

from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class GetDataOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.GET_DATA

    def __init__(self, magic: Optional[int] = None, inv_type: Optional[Union[int, bytes]] = None,
                 block: Optional[OntObjectHash] = None, buf: Optional[bytearray] = None):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_GET_DATA_MSG_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            if isinstance(inv_type, int):
                struct.pack_into("<B", buf, off, inv_type)
            else:
                # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for
                #  2nd param but got `Optional[bytes]`.
                buf[off: off + ont_constants.ONT_CHAR_LEN] = inv_type
            off += ont_constants.ONT_CHAR_LEN
            assert block is not None
            buf[off:off + ont_constants.ONT_HASH_LEN] = block.get_big_endian()
            off += ont_constants.ONT_HASH_LEN

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._inv_type = None
        self._block = None

    def inv_type(self) -> Tuple[int, OntObjectHash]:
        if self._inv_type is None:
            off = ont_constants.ONT_HDR_COMMON_OFF
            self._inv_type, = struct.unpack_from("<B", self.buf, off)
            off += ont_constants.ONT_CHAR_LEN
            self._block = OntObjectHash(buf=self.buf, offset=off, length=ont_constants.ONT_HASH_LEN)

        inv_type, block = self._inv_type, self._block
        assert isinstance(inv_type, int)
        assert isinstance(block, OntObjectHash)
        return inv_type, block



from typing import List, TypeVar, Generic, Iterator

from bxcommon.utils.object_encoder import ObjectEncoder

TItemRaw = TypeVar("TItemRaw")
TItemEncoded = TypeVar("TItemEncoded")


class VectorProxy(Generic[TItemRaw, TItemEncoded]):

    def __init__(
            self, vector: List[TItemRaw], encoder: ObjectEncoder[TItemRaw, TItemEncoded]
    ):
        self.vector = vector
        self._encoder = encoder

    def __repr__(self) -> str:
        return self.vector.__repr__()

    def __str__(self) -> str:
        return self.vector.__str__()

    def __getitem__(self, idx: int) -> TItemEncoded:
        return self._encoder.encode(self.vector[idx])

    def __len__(self) -> int:
        return len(self.vector)

    def __iter__(self) -> Iterator[TItemEncoded]:
        for item in self.vector:
            yield self._encoder.encode(item)

    def __delitem__(self, idx: int) -> None:
        del self.vector[idx]

    def pop(self, idx: int = -1) -> TItemEncoded:
        item = self.vector[idx]
        self.__delitem__(idx)
        return self._encoder.encode(item)

    def append(self, item: TItemEncoded) -> None:
        self.vector.append(self._encoder.decode(item))



from typing import Dict, List, Optional, Tuple


class Flag:
    names_map: Optional[Dict["Flag", str]] = None
    last_used_bitwise_shift = 0

    def __init__(self, bitwise_shift_value: Optional[int] = None, value: Optional[int] = None):
        if bitwise_shift_value:
            self.value = 1 << bitwise_shift_value
        elif value:
            self.value = value
        else:
            self.__class__.last_used_bitwise_shift += 1
            self.value = 1 << self.__class__.last_used_bitwise_shift

    def __hash__(self):
        return hash(self.value)

    def __contains__(self, other: "Flag"):
        return other.value & self.value == other.value

    def __bool__(self):
        return bool(self.value)

    def __or__(self, other):
        return self.__class__(value=self.value | other.value)

    def __and__(self, other):
        return self.__class__(value=self.value & other.value)

    def __xor__(self, other):
        return self.__class__(value=self.value ^ other.value)

    def __invert__(self):
        return self.__class__(value=~self.value)

    def __str__(self):
        flags = self._decompose()
        return "|".join(map(lambda flag: self.names_map[flag], flags))

    def __repr__(self):
        return str(self)

    def _decompose(self) -> List["Flag"]:
        result = []

        names_map = self.names_map
        assert names_map is not None

        for flag in names_map:
            if flag in self:
                result.append(flag)

        return result


class FlagCollection:

    @classmethod
    def init(cls, item_class):
        if item_class.names_map is None:
            item_class.names_map = {}

            for item, name in cls._get_items_names():
                item_class.names_map[item] = name

    @classmethod
    def _get_items_names(cls) -> List[Tuple[Flag, str]]:
        results = []

        for attribute in dir(cls):
            if attribute in cls.__dict__:
                item = cls.__dict__[attribute]

                if isinstance(item, Flag):
                    results.append((item, attribute))

        return results



from typing import Optional

import blxr_rlp as rlp
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxutils.logging.log_level import LogLevel

from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class GetBlockHeadersEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.GET_BLOCK_HEADERS

    fields = [("block_hash", rlp.sedes.binary),
              ("amount", rlp.sedes.big_endian_int),
              ("skip", rlp.sedes.big_endian_int),
              ("reverse", rlp.sedes.big_endian_int)]

    def __repr__(self):
        block_hash = self.get_block_hash()
        if block_hash is None:
            block_requested = f"block_number: {self.get_block_number()}"
        else:
            block_requested = f"block_hash: {block_hash}"
        return f"GetBlockHeadersEthProtocolMessage<{block_requested}, " \
            f"amount: {self.get_amount()}, skip: {self.get_skip()}, reverse: {self.get_reverse()}>"

    def log_level(self):
        return LogLevel.DEBUG

    def get_block_hash(self) -> Optional[Sha256Hash]:
        block_hash_bytes = self.get_field_value("block_hash")
        if len(block_hash_bytes) != eth_common_constants.BLOCK_HASH_LEN:
            return None
        else:
            return Sha256Hash(block_hash_bytes)

    def get_block_number(self) -> Optional[int]:
        block_hash_bytes = self.get_field_value("block_hash")
        if len(block_hash_bytes) == eth_common_constants.BLOCK_HASH_LEN:
            return None
        else:
            return int.from_bytes(block_hash_bytes, byteorder="big")

    def get_amount(self) -> int:
        return self.get_field_value("amount")

    def get_skip(self) -> int:
        return self.get_field_value("skip")

    def get_reverse(self) -> int:
        return self.get_field_value("reverse")



from typing import TYPE_CHECKING

from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.eth.eth_base_connection import EthBaseConnection
from bxgateway.connections.eth.eth_node_connection_protocol import EthNodeConnectionProtocol

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode


class EthNodeConnection(EthBaseConnection):
    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"):
        super().__init__(sock, node)
        self.connection_protocol = EthNodeConnectionProtocol(
            self, self.is_handshake_initiator, self.rlpx_cipher
        )

    def connection_public_key(
        self, sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"
    ) -> bytes:
        peer_ip, peer_port = sock.endpoint
        return node.get_node_public_key(peer_ip, peer_port)

    def get_connection_state_details(self):
        """
        Returns details of the current connection state. Used to submit details to SDN on disconnect.
        :return:
        """
        return f"Connection state: {self.state}. Protocol status: {self.connection_protocol.connection_status}."



import os
from typing import Optional

from bxcommon.models.node_type import NodeType
from bxutils import constants
from bxutils.common import url_helper
from bxutils.ssl.data.ssl_certificate_info import SSLCertificateInfo
from bxutils.ssl.data.ssl_file_info import SSLFileInfo
from bxutils.ssl.data.ssl_storage_info import SSLStorageInfo
from bxutils.ssl.ssl_certificate_type import SSLCertificateType

CA_DIR_NAME: str = SSLCertificateType.CA.name.lower()
REGISTRATION_DIR_NAME: str = SSLCertificateType.REGISTRATION_ONLY.name.lower()
PRIVATE_DIR_NAME: str = SSLCertificateType.PRIVATE.name.lower()


def get_cert_file_name(node_type: NodeType) -> str:
    node_name = node_type.name.lower()
    return constants.SSL_CERT_FILE_FORMAT.format(node_name)


def get_key_file_name(node_type: NodeType) -> str:
    node_name = node_type.name.lower()
    return constants.SSL_KEY_FILE_FORMAT.format(node_name)


def create_storage_info(
    *,
    node_type: NodeType,
    is_ca: bool = False,
    ca_cert_url: Optional[str] = None,
    private_ssl_base_url: Optional[str] = None,
    data_dir: str,
    ca_dir_name: str = CA_DIR_NAME,
    registration_dir_name: str = REGISTRATION_DIR_NAME,
    private_dir_name: str = PRIVATE_DIR_NAME
) -> SSLStorageInfo:
    node_name = node_type.name.lower()
    node_ssl_directory = os.path.join(data_dir, constants.SSL_FOLDER, node_name)

    if ca_cert_url is None or private_ssl_base_url is None:
        ca_cert_url = url_helper.url_join("file:", node_ssl_directory, SSLCertificateType.CA.name.lower())
        private_ssl_base_url = url_helper.url_join("file:", node_ssl_directory)
        node_base_url = private_ssl_base_url
    else:
        node_base_url = url_helper.url_join(private_ssl_base_url, node_name)

    ca_cert_name = constants.SSL_CERT_FILE_FORMAT.format(ca_dir_name)
    if is_ca:
        ca_key_name = constants.SSL_KEY_FILE_FORMAT.format(ca_dir_name)
        ca_key_base_url = url_helper.url_join(private_ssl_base_url, ca_dir_name)
        ca_key_info = SSLFileInfo(
            ca_dir_name,
            ca_key_name,
            url_helper.url_join(ca_key_base_url, ca_key_name)
        )
    else:
        ca_key_info = None

    cert_file_name = get_cert_file_name(node_type)
    key_file_name = get_key_file_name(node_type)
    return SSLStorageInfo(
        node_ssl_directory,
        SSLCertificateInfo(
            SSLFileInfo(
                ca_dir_name,
                ca_cert_name,
                url_helper.url_join(ca_cert_url, ca_cert_name)
            ),
            ca_key_info
        ),
        SSLCertificateInfo(
            SSLFileInfo(
                private_dir_name,
                cert_file_name
            ),
            SSLFileInfo(
                private_dir_name,
                key_file_name
            )
        ),
        SSLCertificateInfo(
            SSLFileInfo(
                registration_dir_name,
                cert_file_name,
                url_helper.url_join(node_base_url, registration_dir_name, cert_file_name)
            ),
            SSLFileInfo(
                registration_dir_name,
                key_file_name,
                url_helper.url_join(node_base_url, registration_dir_name, key_file_name)
            )
        )
    )



import asyncio
import os
import time
from abc import ABCMeta, abstractmethod
from asyncio import Future
from collections import defaultdict, Counter
from ssl import SSLContext
from typing import List, Optional, Tuple, Dict, NamedTuple, Union, Set

import gc

from bxcommon import constants
from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.connection_pool import ConnectionPool
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.exceptions import TerminationError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.models.authenticated_peer_info import AuthenticatedPeerInfo
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.node_model import NodeModel
from bxcommon.models.node_type import NodeType
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.peer_info import ConnectionPeerInfo
from bxcommon.network.socket_connection_state import SocketConnectionStates
from bxcommon.services import sdn_http_service
from bxcommon.services.broadcast_service import BroadcastService, \
    BroadcastOptions
from bxcommon.services.threaded_request_service import ThreadedRequestService
from bxcommon.services.transaction_service import TransactionService
from bxcommon.storage.serialized_message_cache import SerializedMessageCache
from bxcommon.utils import memory_utils, convert, performance_utils
from bxcommon.utils.alarm_queue import AlarmQueue, AlarmId
from bxcommon.utils.blockchain_utils import bdn_tx_to_bx_tx
from bxcommon.common_opts import CommonOpts
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.memory_statistics_service import memory_statistics
from bxcommon.utils.stats.node_info_service import node_info_statistics
from bxcommon.utils.stats.node_statistics_service import node_stats_service
from bxcommon.utils.stats.throughput_service import throughput_statistics
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxcommon.utils.transaction_short_id_buckets import TransactionShortIdBuckets
from bxutils import log_messages, utils
from bxutils import logging
from bxutils.exceptions.connection_authentication_error import \
    ConnectionAuthenticationError
from bxutils.logging import LogRecordType, LogLevel
from bxutils.services.node_ssl_service import NodeSSLService
from bxutils.ssl.extensions import extensions_factory
from bxutils.ssl.ssl_certificate_type import SSLCertificateType


logger = logging.get_logger(__name__)
memory_logger = logging.get_logger(LogRecordType.BxMemory, __name__)
performance_troubleshooting_logger = logging.get_logger(LogRecordType.PerformanceTroubleshooting, __name__)


class DisconnectRequest(NamedTuple):
    file_no: int
    should_retry: bool


# pylint: disable=too-many-public-methods
class AbstractNode:
    __meta__ = ABCMeta
    FLUSH_SEND_BUFFERS_INTERVAL = constants.OUTPUT_BUFFER_BATCH_MAX_HOLD_TIME * 2
    NODE_TYPE: Optional[NodeType] = None

    def __init__(
        self,
        opts: CommonOpts,
        node_ssl_service: NodeSSLService,
        connection_pool: Optional[ConnectionPool] = None
    ):
        # Event handling queue for delayed events
        self.alarm_queue = AlarmQueue()
        self.node_ssl_service = node_ssl_service
        logger.debug("Initializing node of type: {}", self.NODE_TYPE)
        self.server_endpoints = [
            IpEndpoint(constants.LISTEN_ON_IP_ADDRESS, opts.external_port),
            # TODO: remove this after v1 is no longer supported
            IpEndpoint(constants.LISTEN_ON_IP_ADDRESS, opts.non_ssl_port)
        ]

        self.set_node_config_opts_from_sdn(opts)
        self.opts: CommonOpts = opts
        self.pending_connection_requests: Set[ConnectionPeerInfo] = set()
        self.pending_connection_attempts: Set[ConnectionPeerInfo] = set()
        self.recent_connections: ExpiringDict[str, int] = ExpiringDict(
            self.alarm_queue,
            constants.THROTTLE_RECONNECT_TIME_S,
            name="recent_connections"
        )
        self.outbound_peers: Set[OutboundPeerModel] = opts.outbound_peers.copy()

        if connection_pool is not None:
            self.connection_pool = connection_pool
        else:
            self.connection_pool = ConnectionPool()

        self.should_force_exit = False
        self.should_restart_on_high_memory = False

        self.num_retries_by_ip: Dict[Tuple[str, int], int] = defaultdict(int)

        self.init_node_status_logging()
        self.init_throughput_logging()
        self.init_node_info_logging()
        self.init_memory_stats_logging()
        self.init_block_stats_logging()
        self.init_tx_stats_logging()

        # TODO: clean this up alongside outputbuffer holding time
        # this is Nagle's algorithm and we need to implement it properly
        # flush buffers regularly because of output buffer holding time
        self.alarm_queue.register_approx_alarm(self.FLUSH_SEND_BUFFERS_INTERVAL,
                                               constants.OUTPUT_BUFFER_BATCH_MAX_HOLD_TIME,
                                               self.flush_all_send_buffers)

        self.network_num = opts.blockchain_network_num
        self.broadcast_service = self.get_broadcast_service()

        # converting setting in MB to bytes
        self.next_report_mem_usage_bytes = self.opts.dump_detailed_report_at_memory_usage * 1024 * 1024

        if opts.dump_removed_short_ids:
            os.makedirs(opts.dump_removed_short_ids_path, exist_ok=True)

        # each time a network has an update regarding txs, blocks, etc. register in a dict,
        # this way can verify if node lost connection to requested relay.

        self.last_sync_message_received_by_network: Dict[int, float] = {}

        self.start_sync_time: Optional[float] = None
        self.sync_metrics: Dict[int, Counter] = defaultdict(Counter)
        self.sync_short_id_buckets: Dict[int, TransactionShortIdBuckets] = defaultdict(TransactionShortIdBuckets)

        opts.has_fully_updated_tx_service = False

        self.check_sync_relay_connections_alarm_id: Optional[AlarmId] = None
        self.transaction_sync_timeout_alarm_id: Optional[AlarmId] = None

        self.requester = ThreadedRequestService(
            # pyre-fixme[16]: `Optional` has no attribute `name`.
            self.NODE_TYPE.name.lower(), self.alarm_queue, constants.THREADED_HTTP_POOL_SLEEP_INTERVAL_S
        )

        self._last_responsiveness_check_log_time = time.time()
        self._last_responsiveness_check_details = {}
        self.gc_logging_enabled = False
        self.serialized_message_cache = SerializedMessageCache(self.alarm_queue)

        self.alarm_queue.register_alarm(constants.RESPONSIVENESS_CHECK_INTERVAL_S, self._responsiveness_check_log)

    def get_sdn_address(self):
        """
        Placeholder for net event loop to get the sdn address (relay only).
        :return:
        """
        return

    @abstractmethod
    def get_tx_service(self, network_num: Optional[int] = None) -> TransactionService:
        pass

    @abstractmethod
    def get_outbound_peer_info(self) -> List[ConnectionPeerInfo]:
        pass

    @abstractmethod
    def get_broadcast_service(self) -> BroadcastService:
        pass

    def sync_and_send_request_for_relay_peers(self, network_num: int) -> int:
        """
        Requests potential relay peers from SDN. Merges list with provided command line relays.

        This function retrieves from the SDN potential_relay_peers_by_network
        Then it try to ping for each relay (timeout of 2 seconds). The ping is done in parallel
        Once there are ping result, it calculate the best relay and decides if need to switch relays

        The above can take time, so the functions is split into several internal functions and use the thread pool
        not to block the main thread.
        """

        self.requester.send_threaded_request(
            sdn_http_service.fetch_potential_relay_peers_by_network,
            self.opts.node_id,
            network_num,
            # pyre-fixme[6]: Expected `Optional[Callable[[Future[Any]], Any]]` for 4th parameter `done_callback`
            #  to call `send_threaded_request` but got `BoundMethod[Callable(_process_blockchain_network_from_sdn)
            #  [[Named(self, AbstractRelayConnection), Named(get_blockchain_network_future, Future[Any])], Any],
            #  AbstractRelayConnection]`.
            done_callback=self.process_potential_relays_from_sdn
        )

        return constants.CANCEL_ALARMS

    def process_potential_relays_from_sdn(self, get_potential_relays_future: Future):
        pass

    @abstractmethod
    def build_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> Optional[AbstractConnection]:
        pass

    @abstractmethod
    def on_failed_connection_retry(
        self, ip: str, port: int, connection_type: ConnectionType, connection_state: ConnectionState
    ) -> None:
        pass

    def report_connection_attempt(self, ip_address: str) -> int:
        if ip_address in self.recent_connections:
            self.recent_connections[ip_address] += 1
        else:
            self.recent_connections.add(ip_address, 1)
        return self.recent_connections[ip_address]

    def connection_exists(
        self,
        ip: str,
        port: int,
        peer_id: Optional[str] = None
    ) -> bool:
        return self.connection_pool.has_connection(ip, port, peer_id)

    def on_connection_added(self, socket_connection: AbstractSocketConnectionProtocol) -> Optional[AbstractConnection]:
        """
        Notifies the node that a connection is coming in.
        """
        # If we're already connected to the remote peer, log the event and request disconnect.
        self.pending_connection_attempts.discard(
            ConnectionPeerInfo(socket_connection.endpoint, AbstractConnection.CONNECTION_TYPE)
        )
        ip, port = socket_connection.endpoint
        peer_info: Optional[AuthenticatedPeerInfo] = None
        if socket_connection.is_ssl:
            try:
                peer_info = self._get_socket_peer_info(
                    socket_connection
                )
            except ConnectionAuthenticationError as e:
                logger.warning(log_messages.FAILED_TO_AUTHENTICATE_CONNECTION, ip, port, e)
                socket_connection.mark_for_close(should_retry=False)
                return None

            if self.connection_exists(ip, port, peer_info.peer_id)\
                    and peer_info.connection_type != ConnectionType.RELAY_PROXY:
                logger.debug(
                    "Duplicate connection attempted to: {}:{} (peer id: {}). "
                    "Dropping.",
                    ip,
                    port,
                    peer_info.peer_id
                )
                socket_connection.mark_for_close(should_retry=False)
                return None
        elif self.connection_exists(ip, port):
            logger.debug(
                "Duplicate connection attempt to {}:{}. Dropping.",
                ip,
                port,
            )
            socket_connection.mark_for_close(should_retry=False)
            return None

        connection = self._initialize_connection(socket_connection)
        if connection is None:
            return None

        if peer_info is not None:
            connection.on_connection_authenticated(
                peer_info
            )
            self.connection_pool.index_conn_node_id(
                peer_info.peer_id, connection
            )

        connection.state |= ConnectionState.INITIALIZED
        logger.debug("Successfully initialized connection: {}", connection)
        return connection

    def on_connection_closed(self, file_no: int, mark_connection_for_close: bool = False):
        conn = self.connection_pool.get_by_fileno(file_no)

        if conn is None:
            logger.debug("Unexpectedly closed connection not in pool. file_no: {}", file_no)
            return

        if mark_connection_for_close:
            conn.mark_for_close()

        self._destroy_conn(conn)

    def log_refused_connection(self, peer_info: ConnectionPeerInfo, error: str):
        logger.info("Failed to connect to: {}, {}.", peer_info, error)

    def log_closed_connection(self, connection: AbstractConnection):
        if not connection.established:
            logger.info("Failed to connect to: {}.", connection)
        else:
            logger.info("Closed connection: {}", connection)

    def on_updated_peers(self, outbound_peer_models: Set[OutboundPeerModel]) -> None:
        if not outbound_peer_models:
            logger.debug("Got peer update with no peers.")
            return

        logger.debug("Processing updated outbound peers: {}.", outbound_peer_models)

        # Remove peers not in updated list or from command-line args.
        old_peers = self.outbound_peers

        # TODO: remove casting to set once the type of outbound peer model is verified globally
        remove_peers = set(old_peers) - set(outbound_peer_models) - set(self.opts.outbound_peers)

        for rem_peer in remove_peers:
            if self.connection_pool.has_connection(
                rem_peer.ip,
                rem_peer.port,
                rem_peer.node_id
            ):
                rem_conn = self.connection_pool.get_by_ipport(
                    rem_peer.ip, rem_peer.port, rem_peer.node_id
                )
                if rem_conn:
                    rem_conn.mark_for_close(False)

        # Connect to peers not in our known pool or in opts.outbound_peers
        new_peer_models = set()
        new_peer_models.update(old_peers)
        new_peer_models.update(outbound_peer_models)
        new_peer_models.difference_update(remove_peers)
        for peer in new_peer_models:
            peer_ip = peer.ip
            peer_port = peer.port
            if self.should_connect_to_new_outbound_peer(peer):
                node_type = self.NODE_TYPE
                assert node_type is not None
                self.enqueue_connection(
                    peer_ip, peer_port, convert.peer_node_to_connection_type(
                        node_type, peer.node_type
                    )
                )
        self.outbound_peers = new_peer_models

    def on_updated_node_model(self, new_node_model: NodeModel):
        """
        Updates `opts` according a newly updated `NodeModel`.
        This is currently unused on gateways.
        """
        logger.debug(
            "Updating node attributes with new model: {}", new_node_model
        )
        for key, val in new_node_model.__dict__.items():
            logger.trace(
                "Updating attribute '{}': {} => {}", key, self.opts.__dict__.get(key, 'None'), val
            )
            self.opts.__dict__[key] = val

    def should_connect_to_new_outbound_peer(self, outbound_peer: OutboundPeerModel) -> bool:
        return not self.connection_pool.has_connection(
            outbound_peer.ip, outbound_peer.port, outbound_peer.node_id
        )

    def on_bytes_received(self, file_no: int, bytes_received: Union[bytearray, bytes]) -> None:
        """
        :param file_no:
        :param bytes_received:
        :return: True if the node should continue receiving bytes from the remote peer. False otherwise.
        """
        conn = self.connection_pool.get_by_fileno(file_no)

        if conn is None:
            logger.debug("Received bytes for connection not in pool. file_no: {0}", file_no)
            return

        if not conn.is_alive():
            conn.log_trace("Skipping receiving bytes for closed connection.")
            return

        conn.add_received_bytes(bytes_received)
        conn.process_message()

    def get_bytes_to_send(self, file_no: int) -> Optional[memoryview]:
        conn = self.connection_pool.get_by_fileno(file_no)

        if conn is None:
            logger.debug("Request to get bytes for connection not in pool. file_no: {}", file_no)
            return None

        if not conn.is_alive():
            conn.log_trace("Skipping sending bytes for closed connection.")
            return None

        return conn.get_bytes_to_send()

    def on_bytes_sent(self, file_no: int, bytes_sent: int):
        conn = self.connection_pool.get_by_fileno(file_no)

        if conn is None:
            logger.debug("Bytes sent call for connection not in pool. file_no: {0}", file_no)
            return

        conn.advance_sent_bytes(bytes_sent)

    def fire_alarms(self) -> float:
        time_to_next = self.alarm_queue.fire_ready_alarms()
        if time_to_next is not None:
            return time_to_next
        else:
            return constants.MAX_EVENT_LOOP_TIMEOUT

    def force_exit(self):
        """
        Indicates if node should trigger exit in event loop. Primarily used for testing.

        Typically requires one additional socket call (e.g. connecting to this node via a socket)
        to finish terminating the event loop.
        """
        return self.should_force_exit

    async def close(self):
        logger.info("Node is closing! Closing everything.")

        shutdown_task = asyncio.ensure_future(self.close_all_connections())
        try:
            await asyncio.wait_for(shutdown_task, constants.NODE_SHUTDOWN_TIMEOUT_S)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception("Node shutdown failed due to an error: {}, force closing!", e)
        self.requester.close()
        self.cleanup_memory_stats_logging()

    async def close_all_connections(self):
        """
        Closes all connections from the node
        """
        for _, conn in self.connection_pool.items():
            conn.mark_for_close(should_retry=False)

    def broadcast(
        self,
        msg: AbstractMessage,
        broadcasting_conn: Optional[AbstractConnection] = None,
        prepend_to_queue: bool = False,
        connection_types: Optional[Tuple[ConnectionType, ...]] = None
    ) -> List[AbstractConnection]:
        """
        Broadcasts message msg to connections of the specified type except requester.
        """
        if connection_types is None:
            connection_types = (ConnectionType.RELAY_ALL,)
        options = BroadcastOptions(broadcasting_conn, prepend_to_queue, connection_types)
        connections = self.broadcast_service.broadcast(msg, options)
        return connections

    def enqueue_connection(self, ip: str, port: int, connection_type: ConnectionType):
        """
        Queues a connection up for the event loop to open a socket for.
        """
        peer_info = ConnectionPeerInfo(IpEndpoint(ip, port), connection_type)
        if peer_info in self.pending_connection_attempts:
            logger.debug("Not adding {}, waiting until connection attempt to complete", peer_info)
        else:
            logger.trace("Enqueuing connection: {}.", peer_info)
            self.pending_connection_requests.add(peer_info)

    def dequeue_connection_requests(self) -> Optional[Set[ConnectionPeerInfo]]:
        """
        Returns the pending connection requests for the event loop to initiate a socket connection to.
        """
        if self.pending_connection_requests:
            pending_connection_requests = self.pending_connection_requests
            self.pending_connection_requests = set()
            self.pending_connection_attempts.update(pending_connection_requests)
            return pending_connection_requests
        else:
            return None

    def continue_retrying_connection(self, ip: str, port: int, connection_type: ConnectionType) -> bool:
        """
        Indicates whether to continue retrying connection. For most connections, this will will stop
        at the maximum of constants.MAX_CONNECT_RETRIES, but some connections should be always retried
        unless there's some fatal socket error.
        """
        is_sdn = ConnectionType.SDN in connection_type
        return is_sdn or self.num_retries_by_ip[(ip, port)] < constants.MAX_CONNECT_RETRIES

    def init_node_status_logging(self):
        node_stats_service.set_node(self)
        self.alarm_queue.register_alarm(constants.FIRST_STATS_INTERVAL_S, node_stats_service.flush_info)

    def init_throughput_logging(self):
        throughput_statistics.set_node(self)
        self.alarm_queue.register_alarm(constants.FIRST_STATS_INTERVAL_S, throughput_statistics.flush_info)

    def init_node_info_logging(self):
        node_info_statistics.set_node(self)
        self.alarm_queue.register_alarm(constants.FIRST_STATS_INTERVAL_S, node_info_statistics.flush_info)

    def cleanup_memory_stats_logging(self):
        memory_statistics.stop_recording()

    def init_block_stats_logging(self):
        block_stats.set_node(self)

    def init_tx_stats_logging(self):
        tx_stats.set_node(self)

    def flush_all_send_buffers(self):
        for conn in self.connection_pool:
            if conn.socket_connection.can_send:
                conn.socket_connection.send()
        return self.FLUSH_SEND_BUFFERS_INTERVAL

    def record_mem_stats(self, low_threshold: int, medium_threshold: int, high_threshold: int):
        """
        When overridden, records identified memory stats and flushes them to std out
        :returns memory stats flush interval
        """
        total_memory = memory_utils.get_app_memory_usage()
        if total_memory > low_threshold:
            gc.collect()
            total_memory = memory_utils.get_app_memory_usage()
        self._record_mem_stats(total_memory > medium_threshold)

        return memory_statistics.flush_info(high_threshold)

    def _record_mem_stats(self, include_data_structure_memory: bool = False):
        if include_data_structure_memory:
            self.connection_pool.log_connection_pool_mem_stats()

    def set_node_config_opts_from_sdn(self, opts: CommonOpts) -> None:
        blockchain_networks: Dict[int, BlockchainNetworkModel] = opts.blockchain_networks
        for blockchain_network in blockchain_networks.values():
            tx_stats.configure_network(
                blockchain_network.network_num,
                blockchain_network.tx_percent_to_log_by_hash,
                blockchain_network.tx_percent_to_log_by_sid
            )
        bdn_tx_to_bx_tx.init(blockchain_networks)

    def dump_memory_usage(self, total_memory: int, threshold: int):
        if total_memory > threshold and logger.isEnabledFor(LogLevel.DEBUG):
            node_size = self.get_node_memory_size()
            memory_logger.debug(
                "Application consumed {} bytes which is over set limit {} bytes. Detailed memory report: {}",
                total_memory,
                threshold,
                node_size
            )

    def get_node_memory_size(self):
        return memory_utils.get_detailed_object_size(self)

    def on_input_received(self, file_no: int) -> bool:
        """handles an input event from the event loop

        :param file_no: the socket connection file_no
        :return: True if the connection is receivable, otherwise False
        """
        connection = self.connection_pool.get_by_fileno(file_no)
        if connection is None:
            return False
        return connection.on_input_received()

    async def init(self) -> None:
        self.requester.start()

    def handle_connection_closed(
        self, should_retry: bool, peer_info: ConnectionPeerInfo, connection_state: ConnectionState
    ) -> None:
        self.pending_connection_attempts.discard(peer_info)
        peer_ip, peer_port = peer_info.endpoint
        connection_type = peer_info.connection_type
        if should_retry and self.continue_retrying_connection(peer_ip, peer_port, connection_type):
            self.alarm_queue.register_alarm(
                self._get_next_retry_timeout(peer_ip, peer_port),
                self._retry_init_client_socket,
                peer_ip, peer_port, connection_type
            )
        else:
            self.on_failed_connection_retry(peer_ip, peer_port, connection_type, connection_state)

    def get_server_ssl_ctx(self) -> SSLContext:
        return self.node_ssl_service.create_ssl_context(SSLCertificateType.PRIVATE)

    def get_target_ssl_ctx(self, endpoint: IpEndpoint, connection_type: ConnectionType) -> SSLContext:
        logger.trace("Fetching SSL certificate for: {} ({}).", endpoint, connection_type)
        return self.node_ssl_service.create_ssl_context(SSLCertificateType.PRIVATE)

    @abstractmethod
    def reevaluate_transaction_streamer_connection(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def on_new_subscriber_request(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def init_memory_stats_logging(self):
        raise NotImplementedError

    @abstractmethod
    def sync_tx_services(self):
        self.start_sync_time = time.time()
        self.sync_metrics = defaultdict(Counter)

    @abstractmethod
    def _transaction_sync_timeout(self) -> int:
        pass

    @abstractmethod
    def check_sync_relay_connections(self, conn: AbstractConnection) -> int:
        pass

    def _get_socket_peer_info(self, sock: AbstractSocketConnectionProtocol) -> AuthenticatedPeerInfo:
        assert sock.is_ssl
        assert self.NODE_TYPE is not None

        cert = sock.get_peer_certificate()
        node_type = extensions_factory.get_node_type(cert)
        try:
            connection_type = convert.peer_node_to_connection_type(
                # pyre-fixme[6]: Expected `NodeType` for 1st param but got
                #  `Optional[NodeType]`.
                self.NODE_TYPE, node_type
            )
        except (KeyError, ValueError):
            raise ConnectionAuthenticationError(
                f"Peer ssl certificate ({cert}) has an invalid node type: {node_type}!"
            )
        peer_id = extensions_factory.get_node_id(cert)
        if peer_id is None:
            raise ConnectionAuthenticationError(
                f"Peer ssl certificate ({cert}) does not contain a node id!")

        account_id = extensions_factory.get_account_id(cert)
        node_privileges = extensions_factory.get_node_privileges(cert)
        return AuthenticatedPeerInfo(connection_type, peer_id, account_id, node_privileges)

    def _should_log_closed_connection(self, _connection: AbstractConnection) -> bool:
        return True

    def _destroy_conn(self, conn: AbstractConnection):
        """
        Clean up the associated connection and update all data structures tracking it.

        Do not call this function directly to close a connection, unless circumstances do not allow cleaning shutting
        down the node via event loop lifecycle hooks (e.g. immediate shutdown).

        In connection handlers, use `AbstractConnection#mark_for_close`, and the connection will be cleaned up as part
        of event handling.
        In other node lifecycle events, use `enqueue_disconnect` to allow the event loop to trigger connection cleanup.

        :param conn connection to destroy
        """

        if self._should_log_closed_connection(conn):
            self.log_closed_connection(conn)

        should_retry = SocketConnectionStates.DO_NOT_RETRY not in conn.socket_connection.state

        logger.debug("Breaking connection to {}. Attempting retry: {}", conn, should_retry)

        self.connection_pool.delete(conn)
        self.handle_connection_closed(
            should_retry, ConnectionPeerInfo(conn.endpoint, conn.CONNECTION_TYPE), conn.state
        )
        conn.dispose()

    def _initialize_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> Optional[AbstractConnection]:
        conn_obj = self.build_connection(socket_connection)
        ip, port = socket_connection.endpoint
        if conn_obj is not None:
            logger.debug("Connecting to: {}...", conn_obj)

            self.alarm_queue.register_alarm(constants.CONNECTION_TIMEOUT, self._connection_timeout, conn_obj)
            self.connection_pool.add(socket_connection.file_no, ip, port, conn_obj)

            if conn_obj.CONNECTION_TYPE == ConnectionType.SDN:
                # pyre-fixme[16]: `AbstractNode` has no attribute `sdn_connection`.
                self.sdn_connection = conn_obj
        else:
            logger.warning(log_messages.UNABLE_TO_DETERMINE_CONNECTION_TYPE, ip, port)
            socket_connection.mark_for_close(should_retry=False)

        return conn_obj

    def on_network_synced(self, network_num: int) -> None:
        if network_num in self.last_sync_message_received_by_network:
            del self.last_sync_message_received_by_network[network_num]

    def on_fully_updated_tx_service(self):
        logger.debug(
            "Synced transaction state with BDN, last_sync_message_received_by_network: {}",
            self.last_sync_message_received_by_network
        )
        self.opts.has_fully_updated_tx_service = True

    def _connection_timeout(self, conn: AbstractConnection) -> int:
        """
        Check if the connection is established.
        If it is not established, we give up for untrusted connections and try again for trusted connections.
        """

        logger.trace("Checking connection status: {}", conn)

        if conn.established:
            logger.trace("Connection is still established: {}", conn)
            self.num_retries_by_ip[(conn.peer_ip, conn.peer_port)] = 0
            return constants.CANCEL_ALARMS

        if not conn.is_alive():
            logger.trace("Connection has already been marked for close: {}", conn)
            return constants.CANCEL_ALARMS

        # Clean up the old connection and retry it if it is trusted
        logger.trace("Connection has timed out: {}", conn)
        conn.mark_for_close()

        # It is connect_to_address's job to schedule this function.
        return constants.CANCEL_ALARMS

    def _kill_node(self, _signum, _stack):
        """
        Kills the node immediately
        """
        self.should_force_exit = True
        raise TerminationError("Node killed.")

    def _get_next_retry_timeout(self, ip: str, port: int) -> int:
        """
        Returns Fibonnaci(n), where n is the number of retry attempts + 1, up to max of Fibonacci(8) == 13.
        """
        sequence_number = min(self.num_retries_by_ip[(ip, port)] + 1, constants.MAX_CONNECT_TIMEOUT_INCREASE)
        return utils.fibonacci(sequence_number)

    def _retry_init_client_socket(self, ip: str, port: int, connection_type: ConnectionType):
        self.num_retries_by_ip[(ip, port)] += 1

        logger.debug("Retrying {} connection to {}:{}. Attempt #{}.", connection_type, ip, port,
                     self.num_retries_by_ip[(ip, port)])
        self.enqueue_connection(ip, port, connection_type)

        return 0

    def _responsiveness_check_log(self):
        details = ""
        if self.gc_logging_enabled:
            gen0_stats, gen1_stats, gen2_stats = gc.get_stats()

            last_gen0_collections = self._last_responsiveness_check_details.get(
                "gen0_collections", 0
            )
            last_gen1_collections = self._last_responsiveness_check_details.get(
                "gen1_collections", 0
            )
            last_gen2_collections = self._last_responsiveness_check_details.get(
                "gen2_collections", 0
            )

            gen0_diff = gen0_stats["collections"] - last_gen0_collections
            gen1_diff = gen1_stats["collections"] - last_gen1_collections
            gen2_diff = gen2_stats["collections"] - last_gen2_collections

            details = (
                f"gen0_collections: {gen0_diff}, gen1_collections: {gen1_diff}, "
                f"gen2_collections: {gen2_diff}"
            )
            self._last_responsiveness_check_details.update({
                "gen0_collections": gen0_stats["collections"],
                "gen1_collections": gen1_stats["collections"],
                "gen2_collections": gen2_stats["collections"],
            })

        performance_utils.log_operation_duration(
            performance_troubleshooting_logger,
            "Responsiveness Check",
            self._last_responsiveness_check_log_time,
            constants.RESPONSIVENESS_CHECK_INTERVAL_S + constants.RESPONSIVENESS_CHECK_DELAY_WARN_THRESHOLD_S,
            details=details
        )
        self._last_responsiveness_check_log_time = time.time()
        return constants.RESPONSIVENESS_CHECK_INTERVAL_S



import blxr_rlp as rlp
from bxcommon import constants
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class HelloEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.HELLO

    fields = [("version", rlp.sedes.big_endian_int),
              ("client_version_string", rlp.sedes.binary),
              ("capabilities", rlp.sedes.CountableList(rlp.sedes.List([rlp.sedes.binary, rlp.sedes.big_endian_int]))),
              ("listen_port", rlp.sedes.big_endian_int),
              ("remote_pubkey", rlp.sedes.binary)]

    def get_version(self):
        return self.get_field_value("version")

    def get_client_version_string(self):
        return self.get_field_value("client_version_string").decode(constants.DEFAULT_TEXT_ENCODING)

    def get_capabilities(self):
        return self.get_field_value("capabilities")

    def get_listen_port(self):
        return self.get_field_value("listen_port")

    def get_remote_pubkey(self):
        return self.get_field_value("remote_pubkey")



from typing import Iterable
from bxcommon.models.serializable_flag import SerializableFlag


class TxValidationStatus(SerializableFlag):
    """
    Flag used to track transaction validation

    IMPORTANT: Members with explicit values are also defined in extensions.
               Need to be changed in both places if changed here.
    """

    VALID_TX = 1 << 0
    INVALID_FORMAT = 1 << 1
    INVALID_SIGNATURE = 1 << 2
    LOW_FEE = 1 << 3
    REUSE_SENDER_NONCE = 1 << 4

    def __str__(self) -> str:
        # pylint: disable=using-constant-test
        if self.name:
            return str(self.name)
        else:
            return super().__str__()

    def get_subtypes(self) -> Iterable["TxValidationStatus"]:
        for subtype in self.__class__:
            if subtype in self:
                yield subtype



from bxutils.logging_messages_utils import LogMessage
from bxutils.log_message_categories import REQUEST_RESPONSE_CATEGORY, CONNECTION_PROBLEM_CATEGORY, \
    NETWORK_CATEGORY, MEMORY_CATEGORY, \
    GENERAL_CATEGORY, PROCESSING_FAILED_CATEGORY, LOGGING_CATEGORY, \
    NOTIFICATION_FROM_RELAY_CATEGORY, AUTHENTICATION_ERROR, RPC_ERROR

NO_GATEWAY_PEERS = LogMessage(
    "G-000000",
    REQUEST_RESPONSE_CATEGORY,
    "Did not receive expected gateway peers from BDN. Retrying."
)
CLOSE_CONNECTION_ATTEMPT = LogMessage(
    "G-000001",
    CONNECTION_PROBLEM_CATEGORY,
    "Detected attempt to close node connection when another is already established. "
)
ABANDON_GATEWAY_PEER_REQUEST = LogMessage(
    "G-000002",
    NETWORK_CATEGORY,
    "Giving up on querying for gateway peers from the BDN."
)
TX_SYNC_TIMEOUT = LogMessage(
    "G-000003",
    CONNECTION_PROBLEM_CATEGORY,
    "Gateway transaction sync took too long; marking gateway as synced."
)
RELAY_CONNECTION_TIMEOUT = LogMessage(
    "G-000004",
    CONNECTION_PROBLEM_CATEGORY,
    "It has been more than {} seconds since the last time gateway received a message from requested "
)
BLOCKCHAIN_PEER_LACKS_PUBLIC_KEY = LogMessage(
    "G-000005",
    GENERAL_CATEGORY,
    "Received remote blockchain peer without remote public key. This is currently unsupported."
)
BLOCK_COUNT_MISMATCH = LogMessage(
    "G-000006",
    REQUEST_RESPONSE_CATEGORY,
    "Number of blocks received from remote blockchain node ({}) does not match expected ({}). "
    "Temporarily suspending logging of remote blocks sync."
)
INVALID_CONFIG_PAYLOAD = LogMessage(
    "G-000007",
    PROCESSING_FAILED_CATEGORY,
    "Payload in gateway config file ({}) was invalid."
)
READ_CONFIG_FAIL = LogMessage(
    "G-000008",
    PROCESSING_FAILED_CATEGORY,
    "Could not read contents of gateway config file ({}): {}"
)
UNEXPECTED_BLOCKS = LogMessage(
    "G-000009",
    NETWORK_CATEGORY,
    "Received blocks from remote node but nothing is expected."
)
INVALID_CONFIG_LOG_LEVEL = LogMessage(
    "G-000010",
    LOGGING_CATEGORY,
    "Invalid log level provided in config: {}."
)
BLOCK_RECOVERY_NEEDED = LogMessage(
    "G-000011",
    GENERAL_CATEGORY,
    "Block recovery needed. Missing {} sids, {} tx hashes. Total txs in bx_block: {}"
)
UNKNOWN_CALLBACK_MAPPING_WITH_SYNC_REQ = LogMessage(
    "G-000012",
    NETWORK_CATEGORY,
    "Received blockchain sync request type with unknown callback mapping: {}."
)
UNSOLICITED_SYNC_RESPONSE = LogMessage(
    "G-000013",
    REQUEST_RESPONSE_CATEGORY,
    "Received blockchain sync response type that was not asked for or already processed: {}"
)
UNKNOWN_SHORT_IDS = LogMessage(
    "G-000014",
    GENERAL_CATEGORY,
    "Compact block was parsed with {} unknown short ids. Requesting unknown transactions."
)
COMPACT_BLOCK_RECOVERY_FAIL = LogMessage(
    "G-000015",
    PROCESSING_FAILED_CATEGORY,
    "Failed to recover compact block '{}' after receiving BLOCK_TRANSACTIONS message. "
    "Requesting full block."
)
COMPACT_BLOCK_PROCESSING_FAIL = LogMessage(
    "G-000016",
    PROCESSING_FAILED_CATEGORY,
    "Failed to process compact block '{}' after recovery. Requesting full block. Error: {}"
)
STATUS_FILE_JSON_LOAD_FAIL = LogMessage(
    "G-000017",
    GENERAL_CATEGORY,
    "The status file may have been modified. This message can be safely ignored when "
    "running multiple nodes on the same machine. "
    "For more information, please check the status file: {}"
)
REDUNDANT_BLOCK_BODY = LogMessage(
    "G-000018",
    GENERAL_CATEGORY,
    "Block body for hash {} is not in the list of pending new blocks and not marked for cleanup."
)
NO_ACTIVE_BDN_CONNECTIONS = LogMessage(
    "G-000019",
    CONNECTION_PROBLEM_CATEGORY,
    "Gateway does not have an active connection to the relay network. "
    "There may be issues with the BDN. Exiting."
)
UNEXPECTED_TXS_ON_NON_RELAY_CONN = LogMessage(
    "G-000020",
    NETWORK_CATEGORY,
    "Received unexpected txs message on non-tx relay connection: {}"
)
UNEXPECTED_BLOCK_ON_NON_RELAY_CONN = LogMessage(
    "G-000021",
    NETWORK_CATEGORY,
    "Received unexpected block message on non-block relay connection: {}"
)
RPC_INITIALIZATION_FAIL = LogMessage(
    "G-000022",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to initialize Gateway RPC server: {}."
)
RPC_CLOSE_FAIL = LogMessage(
    "G-000023",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to close Gateway RPC server: {}."
)
INVALID_ACCOUNT_ID = LogMessage(
    "G-000024",
    GENERAL_CATEGORY,
    "Valid Account Id is required in order to propagate Transactions as Paid, revert quota type to Free"
)
NO_ACTIVE_BLOCKCHAIN_CONNECTION = LogMessage(
    "G-000025",
    CONNECTION_PROBLEM_CATEGORY,
    "Gateway does not have an active connection to the blockchain node. "
    "Check that the blockchain node is running and available. Exiting."
)
SHORT_ID_RECOVERY_FAIL = LogMessage(
    "G-000026",
    PROCESSING_FAILED_CATEGORY,
    "Could not decompress block {} after attempts to recover short ids. Discarding."
)
PAYLOAD_LENGTH_MISMATCH = LogMessage(
    "G-000027",
    GENERAL_CATEGORY,
    "Payload length does not match buffer size: {} vs {} bytes"
)
PACKET_CHECKSUM_MISMATCH = LogMessage(
    "G-000028",
    GENERAL_CATEGORY,
    "Checksum ({}) for packet doesn't match ({}): {}"
)
BLOCK_COMPRESSION_FAIL = LogMessage(
    "G-000029",
    PROCESSING_FAILED_CATEGORY,
    "Failed to compress block {} - {}"
)
MALFORMED_TX_FROM_RELAY = LogMessage(
    "G-000030",
    NETWORK_CATEGORY,
    "Received malformed transaction message from the BDN."
    "Expected hash: {}. Actual: {}"
)
SOCKET_BUFFER_SEND_FAIL = LogMessage(
    "G-000031",
    NETWORK_CATEGORY,
    "Could not set socket send buffer size: {}"
)
PONG_MESSAGE_TIMEOUT = LogMessage(
    "G-000032",
    CONNECTION_PROBLEM_CATEGORY,
    "Pong message was not received within allocated timeout connection. Closing."
)
_BLOCK_REQUIRES_RECOVERY = LogMessage(
    "G-000033",
    GENERAL_CATEGORY,
    "Block {} requires short id recovery. Querying BDN..."
)
BLOCK_DECOMPRESSION_FAILURE = LogMessage(
    "G-000034",
    GENERAL_CATEGORY,
    "Unexpectedly, could not decompress block {} after block was recovered."
)
BLOCK_WITH_INCONSISTENT_HASHES = LogMessage(
    "G-000035",
    NETWORK_CATEGORY,
    "Received a block with inconsistent hashes from the BDN. "
    "Expected: {}. Actual: {}. Dropping."
)
LACK_BLOCKCHAIN_CONNECTION = LogMessage(
    "G-000036",
    CONNECTION_PROBLEM_CATEGORY,
    "Discarding block. No connection currently exists to the blockchain node."
)
MISMATCH_IP_IN_HELLO_MSG = LogMessage(
    "G-000037",
    REQUEST_RESPONSE_CATEGORY,
    "Received hello message with mismatched IP address. Continuing. Expected ip: {} Socket ip: {}"
)
FAILED_TO_DECOMPRESS_BLOCK = LogMessage(
    "G-000038",
    GENERAL_CATEGORY,
    "Failed to decompress block {} - {}"
)
REDUNDANT_CONNECTION = LogMessage(
    "G-000039",
    CONNECTION_PROBLEM_CATEGORY,
    "Connection already exists, with lower priority. Dropping connection {} and keeping {}."
)
HELLO_MSG_WITH_NO_PEER_ID = LogMessage(
    "G-000040",
    NETWORK_CATEGORY,
    "Received hello message without peer id."
)
PROCESS_BLOCK_FAILURE = LogMessage(
    "G-000041",
    PROCESSING_FAILED_CATEGORY,
    "Failed to process compact block: {}. Request full block. Error: {}"
)
NETWORK_NUMBER_MISMATCH = LogMessage(
    "G-000042",
    PROCESSING_FAILED_CATEGORY,
    "Network number mismatch. Current network num {}, remote network num {}. Closing connection."
)
UNEXPECTED_KEY_MESSAGE = LogMessage(
    "G-000043",
    PROCESSING_FAILED_CATEGORY,
    "Received unexpected key message on non-block relay connection: {}"
)
UNEXPECTED_TX_MESSAGE = LogMessage(
    "G-000044",
    PROCESSING_FAILED_CATEGORY,
    "Received unexpected tx message on non-tx relay connection: {}"
)
SET_SOCKET_BUFFER_SIZE = LogMessage(
    "G-000045",
    NETWORK_CATEGORY,
    "Socket buffer size set was unsuccessful, and was instead set to {}. Reverting to: {}"
)
NOTIFICATION_FROM_RELAY = LogMessage(
    "G-000046",
    NOTIFICATION_FROM_RELAY_CATEGORY,
    "Notification from Relay: {}"
)
WS_INITIALIZATION_FAIL = LogMessage(
    "G-000047",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to initialize Gateway Websockets server: {}."
)
WS_CLOSE_FAIL = LogMessage(
    "G-000048",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to close Gateway Websockets server: {}."
)
ETH_WS_INITIALIZATION_FAIL = LogMessage(
    "G-000049",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not connect to the Ethereum websockets server to verify transactions. "
    "Ensure that you have included the `--ws`, `--ws-api eth`, and `--ws-port PORT` "
    "arguments when starting your Ethereum node, and have specified the right "
    "websocket connection string for `--eth-ws-uri` on your gateway. "
    "If you would like the gateway to run without this functionality, you can remove "
    "the `--eth-ws-uri` argument. Error: {}"
)
ETH_WS_CLOSE_FAIL = LogMessage(
    "G-000050",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to close Etheruem websockets connection: {}."
)
BLOCK_RECOVERY_NEEDED_ONT_CONSENSUS = LogMessage(
    "G-000051",
    GENERAL_CATEGORY,
    "Consensus block recovery needed. Missing {} sids, {} tx hashes. Total txs in bx_block: {}"
)
BLOCK_COMPRESSION_FAIL_ONT_CONSENSUS = LogMessage(
    "G-000052",
    PROCESSING_FAILED_CATEGORY,
    "Failed to compress consensus block {} - {}"
)
_BLOCK_REQUIRES_RECOVERY_ONT_CONSENSUS = LogMessage(
    "G-000053",
    GENERAL_CATEGORY,
    "Consensus block {} requires short id recovery. Querying BDN..."
)
BLOCK_DECOMPRESSION_FAILURE_ONT_CONSENSUS = LogMessage(
    "G-000054",
    GENERAL_CATEGORY,
    "Unexpectedly, could not decompress consensus block {} after block was recovered."
)
LACK_BLOCKCHAIN_CONNECTION_ONT_CONSENSUS = LogMessage(
    "G-000055",
    CONNECTION_PROBLEM_CATEGORY,
    "Discarding consensus block. No connection currently exists to the blockchain node."
)
FAILED_TO_DECOMPRESS_BLOCK_ONT_CONSENSUS = LogMessage(
    "G-000056",
    GENERAL_CATEGORY,
    "Failed to decompress consensus block {} - {}"
)
IPC_INITIALIZATION_FAIL = LogMessage(
    "G-000057",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to initialize Gateway IPC server: {}."
)
IPC_CLOSE_FAIL = LogMessage(
    "G-000058",
    CONNECTION_PROBLEM_CATEGORY,
    "Failed to close Gateway IPC server: {}."
)
TRACKED_BLOCK_CLEANUP_ERROR = LogMessage(
    "G-000059",
    GENERAL_CATEGORY,
    "tracked block cleanup failed, {}"
)
COULD_NOT_DESERIALIZE_TRANSACTION = LogMessage(
    "G-000060",
    PROCESSING_FAILED_CATEGORY,
    "Could not deserialize transaction in transaction service to Ethereum payload: {}, body: {}. Error: {}",
)
TRANSACTION_NOT_FOUND_IN_MEMPOOL = LogMessage(
    "G-000061",
    GENERAL_CATEGORY,
    "Transaction was not found in mempool, despite publication to pendingTxs feed. Publishing "
    "empty contents for {}"
)
ETH_RPC_PROCESSING_ERROR = LogMessage(
    "G-000062",
    PROCESSING_FAILED_CATEGORY,
    "Encountered exception when processing message: {}. Error: {}. Continuing processing.",
)
ETH_RPC_COULD_NOT_RECONNECT = LogMessage(
    "G-000063",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not reconnect to Ethereum websockets feed. Disabling Ethereum transaction "
    "verification for now, but will attempt reconnection upon when the next subscriber "
    "reconnects."
)
WS_COULD_NOT_CONNECT = LogMessage(
    "G-000064",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not connect to websockets server at {}. Connection timed out.",
)
ETH_RPC_ERROR = LogMessage(
    "G-000065",
    PROCESSING_FAILED_CATEGORY,
    "RPC Error response: {}. details: {}.",
)
TRANSACTION_FEED_NOT_ALLOWED = LogMessage(
    "G-000066",
    AUTHENTICATION_ERROR,
    "Account does not have permission to stream transactions. Websocket server will be started "
    "normally, but live transactions will not be available for subscription."
)
ETH_WS_SUBSCRIBER_NOT_STARTED = LogMessage(
    "G-000067",
    GENERAL_CATEGORY,
    "Local Ethereum transaction validation is not enabled. Websocket server will be started "
    "normally, but transaction feed validation will only comes from the BDN and not your connected "
    "Ethereum node, which may provide better performance. Considering starting your gateway with "
    "--eth-ws-uri ws://[ip_address]:[port] for faster transaction streaming."
)
GATEWAY_BAD_FEED_SUBSCRIBER = LogMessage(
    "G-000068",
    GENERAL_CATEGORY,
    "Subscriber {} was not receiving messages and emptying its queue in size of {}. "
    "Disconnecting.",
)
_BAD_RPC_SUBSCRIBER = LogMessage(
    "G-000069",
    RPC_ERROR,
    "Subscription message queue was completed filled up (size {}). "
    "Closing subscription RPC handler and all related subscriptions: {}"
)
ETH_WS_SUBSCRIBER_CONNECTION_BROKEN = LogMessage(
    "G-000070",
    RPC_ERROR,
    "Ethereum websockets connection was broken. Attempting reconnection..."
)
ETH_MISSING_BLOCKCHAIN_IP_AND_BLOCKCHAIN_PEERS = LogMessage(
    "G-000071",
    GENERAL_CATEGORY,
    "At least one of the following four arguments is required for blockchain node connection:"
    "--blockchain-ip, --enode, --blockchain-peers, or --blockchain-peers-file"
)
ETH_MISSING_NODE_PUBLIC_KEY = LogMessage(
    "G-000072",
    GENERAL_CATEGORY,
    "--node-public-key argument is required but not specified."
)
MISSING_BLOCKCHAIN_IP_AND_BLOCKCHAIN_PEERS = LogMessage(
    "G-000073",
    GENERAL_CATEGORY,
    "At least one of the following three arguments is required for blockchain node connection:"
    "--blockchain-ip, --blockchain-peers, or --blockchain-peers-file"
)
MISSING_BLOCKCHAIN_PROTOCOL = LogMessage(
    "G-000074",
    GENERAL_CATEGORY,
    "Blockchain protocol information is missing exiting."
)
INVALID_PUBLIC_KEY_LENGTH = LogMessage(
    "G-000075",
    GENERAL_CATEGORY,
    "Public key must be the 128 digit key associated with the blockchain enode. "
    "Invalid key length: {}"
)
INVALID_PUBLIC_KEY = LogMessage(
    "G-000076",
    GENERAL_CATEGORY,
    "Public key must be constructed from a valid private key."
)
INVALID_BLOCKCHAIN_IP = LogMessage(
    "G-000077",
    GENERAL_CATEGORY,
    "The specified blockchain IP is localhost, which is not compatible with a dockerized "
    "gateway. Did you mean 172.17.0.X?"
)
BLOCKCHAIN_IP_RESOLVE_ERROR = LogMessage(
    "G-000078",
    GENERAL_CATEGORY,
    "Blockchain IP could not be resolved, exiting. Blockchain IP: {}"
)
_COULD_NOT_SERIALIZE_FEED_ENTRY = LogMessage(
    "G-000079",
    PROCESSING_FAILED_CATEGORY,
    "Could not serialize feed entry. Skipping."
)
MISSING_TRANSACTION_STREAMER_PEER_INFO = LogMessage(
    "G-000080",
    CONNECTION_PROBLEM_CATEGORY,
    "Missing transaction streamer peer info upon feed subscription. Connect attempt aborted. Please retry subscription if "
    "you wish to subscribe to the pendingTxs feed and local Ethereum validation is not enabled via --eth-ws-uri."
)
COULD_NOT_DESERIALIZE_BLOCK = LogMessage(
    "G-000081",
    PROCESSING_FAILED_CATEGORY,
    "Could not deserialize block to Ethereum payload: {}, body: {}. Error: {}",
)
BLOCKCHAIN_PROTOCOL_AND_ACCOUNT_MISMATCH = LogMessage(
    "G-000082",
    GENERAL_CATEGORY,
    "Blockchain protocol information does not match account details, exiting."
)
ETH_MISSING_REMOTE_NODE_PUBLIC_KEY = LogMessage(
    "G-000083",
    GENERAL_CATEGORY,
    "--remote-public-key of the blockchain node must be included with command-line specified remote blockchain peer. "
    "Use --remote-public-key."
)
MISSING_BLOCKCHAIN_IP_FROM_BLOCKCHAIN_PEERS = LogMessage(
    "G-000084",
    GENERAL_CATEGORY,
    "blockchain ip is required but not specified in --blockchain-peers argument"
)
ETH_MISSING_BLOCKCHAIN_IP_FROM_BLOCKCHAIN_PEERS = LogMessage(
    "G-000085",
    GENERAL_CATEGORY,
    "Blockchain ip is required but not specified in --blockchain-peers or --blockchain-peers-file"
)
ETH_MISSING_NODE_PUBLIC_KEY_FROM_BLOCKCHAIN_PEERS = LogMessage(
    "G-000086",
    GENERAL_CATEGORY,
    "Node public key is required but not specified in --blockchain-peers or --blockchain-peers-file"
)
ETH_PARSER_INVALID_ENODE_LENGTH = LogMessage(
    "G-000087",
    GENERAL_CATEGORY,
    "Invalid enode: {}, with length: {}"
)
ETH_PARSER_INVALID_ENODE = LogMessage(
    "G-000088",
    GENERAL_CATEGORY,
    "Invalid enode: {}"
)
PARSER_INVALID_PORT = LogMessage(
    "G-000089",
    GENERAL_CATEGORY,
    "Invalid port: {}"
)
BLOCKCHAIN_PEERS_FILE_NOT_FOUND = LogMessage(
    "G-000090",
    GENERAL_CATEGORY,
    "Blockchain peers file not found at path: {}"
)
NODE_EXCEEDS_MEMORY = LogMessage(
    "G-000091",
    MEMORY_CATEGORY,
    "Gateway exceeded allowed memory, restarting"
)
ATTEMPTED_FETCH_FOR_NONEXISTENT_QUEUING_SERVICE = LogMessage(
    "G-000092",
    GENERAL_CATEGORY,
    "Attmepted to fetch queuing service for blockchain node {}, but it was not found."
)



from enum import Enum

CONTENT_TYPE_HEADER_KEY = "Content-Type"
ACCEPT_HEADER_KEY = "Accept"
PLAIN_HEADER_TYPE = "text/plain"
JSON_HEADER_TYPE = "application/json"
OCTET_STREAM_HEADER_TYPE = "application/octet-stream"
FORM_URL_ENCODED_HEADER_TYPE = "application/x-www-form-urlencoded"
TRANSACTION_HASH_PARAMS_KEY = "transaction_hash"
TRANSACTION_PARAMS_KEY = "transaction"
TRANSACTION_VARIATIONS_PARAMS_KEY = "transactions"
TRANSACTION_FEE_BUMP_INTERVAL_S = "bump_interval"
TRANSACTION_JSON_PARAMS_KEY = "transaction_json"
TRANSACTION_FLAG_PARAMS_KEY = "transaction_flag"
PRIVATE_TRANSACTION_TIMEOUT_KEY = "timeout"
BLOCK_NUMBER_PARAMS_KEY = "block_number"
STATE_BLOCK_NUMBER_PARAMS_KEY = "state_block_number"
TIMESTAMP_PARAMS_KEY = "timestamp"
MIN_TIMESTAMP_PARAMS_KEY = "min_timestamp"
MAX_TIMESTAMP_PARAMS_KEY = "max_timestamp"
REVERTING_HASHES_PARAMS_KEY = "reverting_hashes"
WALLET_API_KEY_PARAMS_KEY = "wallet_api_key"
TAG_PARAMS_KEY = "tag"
ALL_KEY = "all"
POOLS_PARAMS_KEY = "pools"
DETAIL_RESPONSE_PARAM_KEY = "detail_response"
SYNCHRONOUS_PARAMS_KEY = "synchronous"
MONITOR_SPEEDUP_PARAMS_KEY = "monitor_speedup"
STATUS_TRACK_PARAMS_KEY = "status_track"
NONCE_MONITORING_PARAMS_KEY = "nonce_monitoring"
DETAILS_LEVEL_PARAMS_KEY = "details_level"
BLOCKCHAIN_PROTOCOL_PARAMS_KEY = "blockchain_protocol"
BLOCKCHAIN_NETWORK_PARAMS_KEY = "blockchain_network"
BLOCKCHAIN_NETWORK_NUM_PARAMS_KEY = "blockchain_network_num"
NETWORK_NUM_PARAMS_KEY = "network_num"
NETWORK_TRANSACTION_FEE_UPDATE_KEY = "network_tx_fee_update"
ALLOWED_TIME_INTERVAL_SAME_NONCE_KEY = "allowed_time_interval_same_nonce"
ALLOWED_NETWORK_TRANSACTION_FEE_CHANGE_SAME_NONCE_KEY = "allowed_network_tx_fee_change_same_nonce"
ACCOUNT_ID_PARAMS_KEY = "account_id"
ACCOUNT_CACHE_KEY_PARAMS_KEY = "account_cache_key"
TX_SERVICE_FILE_NAME_PARAMS_KEY = "file_name"
BLOCKCHAIN_PEER_PARAMS_KEY = "peer"
AUTHORIZATION_HEADER_KEY = "Authorization"
WEBSOCKET_HEADER_KEY = "upgrade"
TRANSACTIONS_PARAMS_KEY = "transactions"
MEV_CREDIT_PARAMS_KEY = "mev_credit"
RPC_SERVER_INIT_TIMEOUT_S = 10
RPC_SERVER_STOP_TIMEOUT_S = 10
HEALTHCHECK_INTERVAL = 60
DEFAULT_RPC_PORT = 28332
DEFAULT_RPC_HOST = "127.0.0.1"
DOCKER_RPC_HOST = "172.17.0.1"
DEFAULT_RPC_USER = ""
DEFAULT_RPC_PASSWORD = ""
DEFAULT_RPC_BASE_SSL_URL = ""
MAINNET_NETWORK_NAME = "Mainnet"
DEFAULT_BLOCKCHAIN_PROTOCOL = "Ethereum"
CLOUD_API_URL = "https://api.blxrbdn.com"

JSON_RPC_VERSION = "2.0"
JSON_RPC_REQUEST_ID = "id"
JSON_RPC_METHOD = "method"
JSON_RPC_PARAMS = "params"
JSON_RPC_VERSION_FIELD = "jsonrpc"
JSON_RPC_RESULT = "result"
JSON_RPC_ERROR = "error"
JSON_RPC_MESSAGE = "message"
JSON_RPC_DATA = "data"
JSON_RPC_ERROR_CODE = "code"

NEW_TRANSACTION_FEED_NAME = "newTxs"
NEW_BLOCKS_FEED_NAME = "newBlocks"
ETH_ON_BLOCK_FEED_NAME = "onBlock"
ETH_PENDING_TRANSACTION_FEED_NAME = "pendingTxs"
TRANSACTION_STATUS_FEED_NAME = "transactionStatus"
ETH_TRANSACTION_RECEIPTS_FEED_NAME = "txReceipts"

ETH_GET_TRANSACTION_RECEIPT_RPC_METHOD = "eth_getTransactionReceipt"
ETH_RPC_INTERNAL_RESPONSE_ITEMS = {"jsonrpc", "id"}


class ContentType(Enum):
    PLAIN = PLAIN_HEADER_TYPE
    JSON = JSON_HEADER_TYPE

    @classmethod
    def from_string(cls, s: str) -> "ContentType":
        for val in cls:
            if val.value == s:
                return val
        raise ValueError(f"{s} is not a valid {cls}.")



import struct
from typing import Optional
from bxgateway import btc_constants
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


class FeeFilterBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.FEE_FILTER

    def __init__(self, magic: Optional[int] = None, buf: Optional[bytearray] = None, fee_rate: Optional[int] = None):
        if buf is None:
            buf = bytearray(btc_constants.BTC_HDR_COMMON_OFF + 8)
            self.buf = buf

            off = btc_constants.BTC_HDR_COMMON_OFF
            struct.pack_into("<Q", buf, off, fee_rate)
            off += 8

            super().__init__(magic, self.MESSAGE_TYPE, off - btc_constants.BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._fee_rate = None

    def fee_rate(self) -> int:
        if self._fee_rate is None:
            self._fee_rate = struct.unpack_from("<Q", self.buf, btc_constants.BTC_HDR_COMMON_OFF)[0]

        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._fee_rate



from typing import List, Union

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxgateway import gateway_constants
from bxgateway.connections.ont.ont_base_connection_protocol import OntBaseConnectionProtocol
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.inventory_ont_message import InvOntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.ver_ack_ont_message import VerAckOntMessage
from bxgateway.messages.ont.version_ont_message import VersionOntMessage


class OntRemoteConnectionProtocol(OntBaseConnectionProtocol):

    def __init__(self, connection):
        super(OntRemoteConnectionProtocol, self).__init__(connection)

        connection.message_handlers.update({
            OntMessageType.VERSION: self.msg_version,
            OntMessageType.BLOCK: self.msg_block,
            OntMessageType.HEADERS: self.msg_proxy_response,
            OntMessageType.INVENTORY: self.msg_inv,
        })
        self.ping_interval_s = gateway_constants.BLOCKCHAIN_PING_INTERVAL_S

    def msg_version(self, _msg: VersionOntMessage):
        """
        Handle version message.
        Gateway initiates connection, so do not check for misbehavior. Record that we received the version message,
        send a verack, and synchronize chains if need be.
        """
        self.connection.on_connection_established()
        reply = VerAckOntMessage(self.magic, self.is_consensus)
        self.connection.enqueue_msg(reply)
        self.connection.node.alarm_queue.register_alarm(self.ping_interval_s,
                                                        self.connection.send_ping)

        if self.connection.is_active():
            self.connection.node.on_remote_blockchain_connection_ready(self.connection)

    def msg_block(self, msg: BlockOntMessage) -> None:
        block_stats.add_block_event_by_block_hash(
            msg.block_hash(),
            BlockStatEventType.REMOTE_BLOCK_RECEIVED_BY_GATEWAY,
            network_num=self.connection.network_num,
            more_info="Protocol: {}, Network: {}".format(
                self.connection.node.opts.blockchain_protocol,
                self.connection.node.opts.blockchain_network
            )
        )
        return self.msg_proxy_response(msg)

    def msg_inv(self, msg: InvOntMessage):
        # Ignoring INV messages from REMOTE NODE to make sure it does not send new blocks to Ontology node
        pass

    def _build_get_blocks_message_for_block_confirmation(
        self, hashes: List[Sha256Hash]
    ) -> AbstractMessage:
        pass

    def _set_transaction_contents(
        self, tx_hash: Sha256Hash, tx_content: Union[memoryview, bytearray]
    ) -> None:
        pass



from abc import abstractmethod, ABCMeta
from asyncio import QueueFull

from typing import TypeVar, Generic, List, Dict, Optional, Any, Set, NamedTuple, Tuple

from bxcommon import log_messages, constants
from bxcommon.feed.subscriber import Subscriber
from bxcommon.feed import filter_parsing
from bxutils import logging

logger = logging.get_logger(__name__)
T = TypeVar("T")
S = TypeVar("S")


class FeedKey(NamedTuple):
    name: str
    network_num: int = 0


class Feed(Generic[T, S], metaclass=ABCMeta):
    FIELDS: List[str] = []
    FILTERS: Set[str] = set()
    ALL_FIELDS: List[str] = []
    name: str
    network_num: int
    subscribers: Dict[str, Subscriber[T]]
    feed_key: FeedKey

    def __init__(self, name: str, network_num: int = constants.ALL_NETWORK_NUM) -> None:
        self.name = name
        self.network_num = network_num
        self.subscribers = {}
        self.feed_key = FeedKey(name, network_num)

    def __repr__(self) -> str:
        return f"Feed<{self.name}>"

    def subscribe(self, options: Dict[str, Any]) -> Subscriber[T]:
        include_fields = options.get("include", "all")
        if include_fields == "all":
            options["include"] = self.ALL_FIELDS

        subscriber: Subscriber[T] = Subscriber(options)
        self.subscribers[subscriber.subscription_id] = subscriber
        return subscriber

    def unsubscribe(self, subscriber_id: str) -> Optional[Subscriber[T]]:
        return self.subscribers.pop(subscriber_id, None)

    def publish(self, raw_message: S) -> None:
        if self.subscriber_count() == 0:
            return

        try:
            serialized_message = self.serialize(raw_message)
        # pylint: disable=broad-except
        except Exception:
            logger.error(log_messages.COULD_NOT_SERIALIZE_FEED_ENTRY, exc_info=True)
            return

        bad_subscribers = []
        cached_subscription_items = {}
        for subscriber in self.subscribers.values():
            if subscriber.should_exit:
                logger.error(
                    log_messages.BAD_FEED_SUBSCRIBER_SHOULD_EXIT, subscriber.subscription_id, self
                )
                bad_subscribers.append(subscriber)
                continue
            if not self.should_publish_message_to_subscriber(
                subscriber, raw_message, serialized_message
            ):
                continue

            try:
                for cached_subscriber, cached_message in cached_subscription_items.items():
                    if subscriber.same_options(cached_subscriber):
                        subscriber.fast_queue(cached_message)
                        break
                else:
                    queued_message = subscriber.queue(serialized_message)
                    cached_subscription_items[subscriber] = queued_message
            except QueueFull:
                logger.info(
                    log_messages.BAD_FEED_SUBSCRIBER, subscriber.subscription_id, self
                )
                bad_subscribers.append(subscriber)

        for bad_subscriber in bad_subscribers:
            self.unsubscribe(bad_subscriber.subscription_id)

    @abstractmethod
    def serialize(self, raw_message: S) -> T:
        """
        Operations to serialize raw data for publication.

        Any potential CPU expensive operations should be moved here, to optimize
        serialization time.
        """
        # pylint: disable=unnecessary-pass
        pass

    def subscriber_count(self) -> int:
        return len(self.subscribers)

    # pylint: disable=unused-argument
    def should_publish_message_to_subscriber(
        self, subscriber: Subscriber[T], raw_message: S, serialized_message: T
    ) -> bool:
        return True

    def validate_filters(self, filters: str) -> Tuple[str, List[str]]:
        filter_parsing.get_validator(filters)
        keys = filter_parsing.get_keys(filters)
        logger.debug("Returning filters {} with keys {}", filters, keys)
        return filters, keys



from typing import Optional

from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils.object_hash import Sha256Hash


class BlockHoldingMessage(AbstractBroadcastMessage):
    """
    Request for other gateways to hold onto the block for a timeout to avoid encrypted block duplication.
    """
    MESSAGE_TYPE = BloxrouteMessageType.BLOCK_HOLDING

    def __init__(self, block_hash: Optional[Sha256Hash] = None, network_num: Optional[int] = None,
                 source_id: str = "", buf: Optional[bytearray] = None):
        self._block_id = None
        super().__init__(block_hash, network_num, source_id, buf)

    def __repr__(self) -> str:
        return f"BlockHoldingMessage<block_hash: {self.block_hash()}>"

    def block_hash(self) -> Sha256Hash:
        return self.message_hash()



import struct

from bxgateway.btc_constants import BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


class SendCompactBtcMessage(BtcMessage):
    """
    Message sent by Bitcoin nodes after initial handshake to notify if the peer would like to know about new blocks
    with Compact Block messages instead of Inv messages
    """

    MESSAGE_TYPE = BtcMessageType.SEND_COMPACT

    ON_FLAG_LEN = 1
    VERSION_LEN = 8

    # pyre-fixme[9]: magic has type `int`; used as `None`.
    # pyre-fixme[9]: on_flag has type `bool`; used as `None`.
    # pyre-fixme[9]: version has type `int`; used as `None`.
    # pyre-fixme[9]: buf has type `memoryview`; used as `None`.
    def __init__(self, magic: int = None, on_flag: bool = None, version: int = None, buf: memoryview = None):
        if buf is None:
            # pyre-fixme[9]: buf has type `memoryview`; used as `bytearray`.
            buf = bytearray(BTC_HDR_COMMON_OFF + self.ON_FLAG_LEN + self.VERSION_LEN)

            off = BTC_HDR_COMMON_OFF

            struct.pack_into("<B", buf, off, on_flag)
            off += self.ON_FLAG_LEN

            struct.pack_into("<Q", buf, off, version)
            off += self.VERSION_LEN

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._memoryview = memoryview(buf)
        self._on_flag = self._version = None

    def on_flag(self) -> bool:
        if self._on_flag is None:
            off = BTC_HDR_COMMON_OFF
            self._on_flag, = struct.unpack_from("<B", self.buf, off)

        # pyre-fixme[7]: Expected `bool` but got `None`.
        return self._on_flag

    def version(self) -> int:
        if self._version is None:
            off = BTC_HDR_COMMON_OFF + self.ON_FLAG_LEN
            self._version, = struct.unpack_from("<Q", self.buf, off)

        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._version


from abc import ABCMeta, abstractmethod
from typing import Set, List, Optional, TYPE_CHECKING

from bxcommon.messages.bloxroute.abstract_cleanup_message import AbstractCleanupMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import Sha256Hash
from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple
from bxcommon import constants
from bxgateway.services.abstract_block_queuing_service import AbstractBlockQueuingService

from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(LogRecordType.BlockCleanup, __name__)


class AbstractBlockCleanupService(SpecialMemoryProperties, metaclass=ABCMeta):
    """
    Service for managing block cleanup.
    """

    def __init__(self, node: "AbstractGatewayNode", network_num: int):
        """
        Constructor
        :param node: reference to node object
        :param network_num: network number
        """

        self.node: "AbstractGatewayNode" = node
        self.network_num: int = network_num

        self._block_hash_marked_for_cleanup: Set[Sha256Hash] = set()
        self.last_confirmed_block: Optional[Sha256Hash] = None

    def is_marked_for_cleanup(self, block_hash: Sha256Hash) -> bool:
        return block_hash in self._block_hash_marked_for_cleanup

    def mark_blocks_and_request_cleanup(self, block_hashes: List[Sha256Hash]) -> None:
        """
        mark already seen blocks sent again from the bitcoin node for cleanup.
        compare the inventory list against our tracked seen blocks in transaction service.
        skipping the # most recent blocks (# refers to block_confirmations_count opt - 1).
        we explicitly request the inventory list from the bitcoin node
        starting from the last block confirmed for cleanup, up to the most recent block
        (or the following 500 in the chain (whichever comes first)).
        the comparison attempts to cross a match between the btc node list and our list,
        starting from the most recent block (skipping the # count), hence the cross_match_idx.
        the cross_match_idx is the index in which we start to mark our tracked blocks for cleanup.
        an example:
        last_confirmed_block = b0
        block_confirmations_count = 6
        block_hashes = [b1, b2, b3, b4, b5, b6, b7, b8, b9]
        tracked_seen_blocks = [b1, b2, b2.1 (fork), b3, b4, b5, b6, b7, b8, b9]
        cross_match_idx = 4
        blocks marked for full cleanup = [b1, b2, b3, b4]
        blocks removed from tracking = [b2.1]
        :param block_hashes: a list of block hashes from a standard inventory message
        :param block_queuing_service: block queuing service for node that triggered cleanup
        """
        skip_block_count = self.node.network.block_confirmations_count
        if len(block_hashes) >= skip_block_count:
            tx_service = self.node.get_tx_service(self.network_num)
            tracked_blocks = tx_service.get_tracked_blocks(0, 0)
            confirmed_blocks = block_hashes[:-skip_block_count]
            unconfirmed_blocks = block_hashes[-skip_block_count:]
            if tracked_blocks:
                logger.debug("Block cleanup flow report: tracked blocks: {}, confirmed blocks: {}, "
                             "unconfirmed blocks: {}", tracked_blocks, confirmed_blocks, unconfirmed_blocks)
                cross_match_idx: Optional[int] = None
                for block_hash in reversed(confirmed_blocks):
                    if block_hash in tracked_blocks:
                        tracked_idx = tracked_blocks[block_hash]
                        if cross_match_idx is None or cross_match_idx < tracked_idx:
                            cross_match_idx = tracked_idx
                        logger.trace("Block cleanup flow requested block: {}", block_hash)
                        self.block_cleanup_request(block_hash)
                        self.node.block_queuing_service_manager.remove(block_hash)
                        del tracked_blocks[block_hash]
                    else:
                        logger.trace("Block cleanup flow confirmed block is not tracked: {}", block_hash)
                if cross_match_idx is not None:
                    for tracked_block, idx in tracked_blocks.items():
                        if idx <= cross_match_idx:
                            logger.debug("Tracked block does not exist: {}", tracked_block)
                            tx_service.on_block_cleaned_up(tracked_block)

    def on_new_block_received(self, block_hash: Sha256Hash, prev_block_hash: Sha256Hash) -> None:
        """
        checks if there are no confirmed blocks for cleanup.
        if True, perform a second check if there are no tracked blocks in tx service
        and add the prev block to the tracker if so.
        this is needed because the request for blocks return only the initial block's children.
        :param block_hash: the current block hash
        :param prev_block_hash: the prev block hash
        """
        if self.last_confirmed_block is None:
            tx_service = self.node.get_tx_service(self.network_num)
            if tx_service.get_tracked_seen_block_count() == 0:
                tx_service.track_seen_short_ids(prev_block_hash, [])
                logger.trace("Setting root block hash for cleanup: first block: {} previous block: {}", block_hash,
                             prev_block_hash)

    @abstractmethod
    def clean_block_transactions(
            self,
            block_msg,
            transaction_service: TransactionService
    ) -> None:
        pass

    def block_cleanup_request(self, block_hash: Sha256Hash) -> None:
        if not self.is_marked_for_cleanup(block_hash):
            self._block_hash_marked_for_cleanup.add(block_hash)
            self.last_confirmed_block = block_hash
            block_queuing_service = None
            if self.node.block_queuing_service_manager.is_in_common_block_storage(block_hash):
                block_queuing_service = self.node.block_queuing_service_manager.get_designated_block_queuing_service()
            if block_queuing_service is not None:
                self.node.alarm_queue.register_alarm(
                    constants.MIN_SLEEP_TIMEOUT,
                    self.clean_block_transactions_from_block_queue,
                    block_hash,
                    block_queuing_service
                )
            elif self.node.has_active_blockchain_peer():
                self._request_block(block_hash)
            else:
                logger.debug("Block cleanup for '{}' failed. No connection to node.", repr(block_hash))

    @abstractmethod
    def clean_block_transactions_from_block_queue(
        self, block_hash: Sha256Hash, block_queuing_service
    ):
        pass

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return super(AbstractBlockCleanupService, self).special_memory_size(ids)

    @abstractmethod
    def contents_cleanup(self,
                         transaction_service: TransactionService,
                         block_confirmation_message: AbstractCleanupMessage
                         ):
        pass

    def process_cleanup_message(
            self,
            msg: AbstractCleanupMessage,
            node: "AbstractGatewayNode",
    ):
        block_cleanup = msg.MESSAGE_TYPE == BloxrouteMessageType.BLOCK_CONFIRMATION
        message_hash = msg.message_hash()
        if not self.node.should_process_block_hash(message_hash):
            return
        transaction_service = node.get_tx_service()
        tracked_blocks = transaction_service.get_tracked_blocks()
        if message_hash in node.block_cleanup_processed_blocks:
            if message_hash in tracked_blocks and block_cleanup:
                logger.debug("Block confirmation message already processed: {}. Reprocessing tracked blocks: {}.",
                             message_hash, tracked_blocks)
            else:
                logger.debug("Cleanup message was already processed. Skipping: {}", message_hash)
                return
        logger.debug("Processing cleanup message: {}", message_hash)
        node.block_cleanup_processed_blocks.add(message_hash)
        self.contents_cleanup(transaction_service, msg)

    @abstractmethod
    def _request_block(self, block_hash: Sha256Hash):
        pass



from typing import Any, Dict, Optional, TYPE_CHECKING

from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.utils.stats import stats_format
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import (
    gateway_bdn_performance_stats_service,
    GatewayBdnPerformanceStatInterval,
)

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

INTERVAL_START_TIME = "interval_start_time"
INTERVAL_END_TIME = "interval_end_time"
BLOCKS_FROM_BDN = "blocks_from_bdn_percentage"
TX_FROM_BDN = "transactions_from_bdn_percentage"
BLOCKS_SEEN = "total_blocks_seen"


class BdnPerformanceRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": "",
        "description": "return percentage of blocks/transactions first received from BDN rather than from p2p network "
        "in the previous 15 minute interval",
    }

    def validate_params(self) -> None:
        pass

    async def process_request(self) -> JsonRpcResponse:
        return self.ok(self._calc_bdn_performance_stats())

    def _calc_bdn_performance_stats(self) -> Dict[str, Any]:
        stats = {}
        interval_data: Optional[
            GatewayBdnPerformanceStatInterval
        ] = gateway_bdn_performance_stats_service.get_most_recent_stats()

        if interval_data is None:
            stats.update({
                INTERVAL_START_TIME: float("nan"),
                INTERVAL_END_TIME: float("nan"),
                BLOCKS_FROM_BDN: float("nan"),
                TX_FROM_BDN: float("nan"),
                BLOCKS_SEEN: float("nan"),
            })
            return stats

        stats[INTERVAL_START_TIME] = str(interval_data.start_time)
        assert interval_data.end_time is not None
        stats[INTERVAL_END_TIME] = str(interval_data.end_time)

        blockchain_node_to_bdn_stats = interval_data.blockchain_node_to_bdn_stats
        assert blockchain_node_to_bdn_stats is not None
        for peer_endpoint, node_stats in blockchain_node_to_bdn_stats.items():
            peer_stats = {}
            if (
                node_stats.new_blocks_received_from_bdn
                + node_stats.new_blocks_received_from_blockchain_node
                == 0
            ):
                peer_stats[BLOCKS_FROM_BDN] = float("nan")
            if (
                node_stats.new_tx_received_from_bdn
                + node_stats.new_tx_received_from_blockchain_node
                == 0
            ):
                peer_stats[TX_FROM_BDN] = float("nan")

            if BLOCKS_FROM_BDN not in peer_stats:
                peer_stats[BLOCKS_FROM_BDN] = stats_format.percentage(
                    100
                    * (
                        node_stats.new_blocks_received_from_bdn
                        / (
                            node_stats.new_blocks_received_from_bdn
                            + node_stats.new_blocks_received_from_blockchain_node
                        )
                    )
                )
            if TX_FROM_BDN not in peer_stats:
                peer_stats[TX_FROM_BDN] = stats_format.percentage(
                    100
                    * (
                        node_stats.new_tx_received_from_bdn
                        / (
                            node_stats.new_tx_received_from_bdn
                            + node_stats.new_tx_received_from_blockchain_node
                        )
                    )
                )
            peer_stats.update({
                BLOCKS_SEEN: node_stats.new_blocks_seen,
            })
            stats[str(peer_endpoint)] = peer_stats

        return stats



from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.keep_alive_message import KeepAliveMessage


class PongMessageV13(KeepAliveMessage):
    MESSAGE_TYPE = BloxrouteMessageType.PONG

    def __init__(self, nonce=None, buf=None) -> None:
        super(PongMessageV13, self).__init__(msg_type=self.MESSAGE_TYPE, nonce=nonce, buf=buf)

    def __repr__(self) -> str:
        return "PongMessage<nonce: {}>".format(self.nonce())



import json
import os
import platform
import sys
from datetime import datetime
from json.decoder import JSONDecodeError
from typing import List, Optional, Set

from prometheus_client import Enum

from bxcommon.connections.connection_pool import ConnectionPool
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.constants import OS_VERSION
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.utils import config
from bxcommon.utils import model_loader
from bxgateway.utils.logging.status.analysis import Analysis
from bxgateway.utils.logging.status.connection_state import ConnectionState
from bxgateway.utils.logging.status.diagnostics import Diagnostics
from bxgateway.utils.logging.status.environment import Environment
from bxgateway.utils.logging.status.extension_modules_state import ExtensionModulesState
from bxgateway.utils.logging.status.gateway_status import GatewayStatus
from bxgateway.utils.logging.status.installation_type import InstallationType
from bxgateway.utils.logging.status.network import Network
from bxgateway.utils.logging.status.summary import Summary
from bxutils import logging
from bxgateway import log_messages
from bxutils.encoding.json_encoder import EnhancedJSONEncoder
from bxgateway.utils.logging.status import summary as summary_status

logger = logging.get_logger(__name__)

STATUS_FILE_NAME = "gateway_status.log"
CONN_TYPES = {ConnectionType.RELAY_ALL, ConnectionType.REMOTE_BLOCKCHAIN_NODE}

gateway_status = Enum(
    "gateway_status",
    "Gateway's online/offline status",
    states=[status.value for status in GatewayStatus]
)


def initialize(use_ext: bool, src_ver: str, ip_address: str, continent: str, country: str,
               update_required: bool, account_id: Optional[str], quota_level: Optional[int]) -> Diagnostics:
    current_time = _get_current_time()
    summary = Summary(
        gateway_status=GatewayStatus.OFFLINE,
        ip_address=ip_address,
        continent=continent,
        country=country,
        update_required=update_required,
        account_info=summary_status.gateway_status_get_account_info(account_id),
        quota_level=summary_status.gateway_status_get_quota_level(quota_level),
    )
    assert summary.gateway_status is not None
    # pyre-fixme[16]: `Optional` has no attribute `value`.
    gateway_status.state(summary.gateway_status.value)
    environment = Environment(_get_installation_type(), OS_VERSION, platform.python_version(), sys.executable)
    network = Network([], [], [])
    analysis = Analysis(current_time, _get_startup_param(), src_ver, _check_extensions_validity(use_ext, src_ver),
                        environment, network, _get_installed_python_modules())
    diagnostics = Diagnostics(summary, analysis)

    _save_status_to_file(diagnostics)
    return diagnostics


def get_diagnostics(use_ext: bool, src_ver: str, ip_address: str, continent: str, country: str,
                    update_required: bool, account_id: Optional[str], quota_level: Optional[int]) -> Diagnostics:
    return _load_status_from_file(use_ext, src_ver, ip_address, continent, country, update_required, account_id, quota_level)


def update(conn_pool: ConnectionPool, use_ext: bool, src_ver: str, ip_address: str, continent: str, country: str,
           update_required: bool, blockchain_peers: Set[BlockchainPeerInfo], account_id: Optional[str],
           quota_level: Optional[int]) -> Diagnostics:
    path = config.get_data_file(STATUS_FILE_NAME)
    if not os.path.exists(path):
        initialize(use_ext, src_ver, ip_address, continent, country, update_required, account_id, quota_level)
    diagnostics = _load_status_from_file(use_ext, src_ver, ip_address, continent, country, update_required, account_id, quota_level)
    analysis = diagnostics.analysis
    network = analysis.network

    for network_type, individual_network in network.iter_network_type_pairs():
        for conn in individual_network:
            if conn.get_connection_state() == ConnectionState.DISCONNECTED or \
                    not conn_pool.has_connection(conn.ip_address, int(conn.port)):
                network.remove_connection(conn, network_type)

    for conn_type in CONN_TYPES:
        for conn in conn_pool.get_by_connection_types([conn_type]):
            network.add_connection(conn.CONNECTION_TYPE, conn.peer_desc, str(conn.file_no), conn.peer_id)

    for blockchain_peer in blockchain_peers:
        blockchain_ip_endpoint = IpEndpoint(blockchain_peer.ip, blockchain_peer.port)
        blockchain_conn = None
        if conn_pool.has_connection(blockchain_peer.ip, blockchain_peer.port):
            blockchain_conn = conn_pool.get_by_ipport(blockchain_peer.ip, blockchain_peer.port)
        if blockchain_conn:
            network.add_connection(ConnectionType.BLOCKCHAIN_NODE, str(blockchain_ip_endpoint),
                                   str(blockchain_conn.file_no), blockchain_conn.peer_id)
        else:
            network.add_connection(ConnectionType.BLOCKCHAIN_NODE, str(blockchain_ip_endpoint), None, None)

    summary = network.get_summary(ip_address, continent, country, update_required, account_id, quota_level)
    assert summary.gateway_status is not None
    # pyre-fixme[16]: `Optional` has no attribute `value`.
    gateway_status.state(summary.gateway_status.value)
    diagnostics = Diagnostics(summary, analysis)

    _save_status_to_file(diagnostics)
    return diagnostics


def update_alarm_callback(conn_pool: ConnectionPool, use_ext: bool, src_ver: str, ip_address: str, continent: str,
                          country: str, update_required: bool, blockchain_peers: Set[BlockchainPeerInfo],
                          account_id: Optional[str], quota_level: Optional[int]) -> None:
    update(conn_pool, use_ext, src_ver, ip_address, continent, country, update_required, blockchain_peers, account_id, quota_level)


def _check_extensions_validity(use_ext: bool, src_ver: str) -> ExtensionModulesState:
    if use_ext:
        try:
            import task_pool_executor as tpe
        except ImportError:
            return ExtensionModulesState.UNAVAILABLE
        extensions_version = tpe.__version__
        if src_ver == extensions_version:
            return ExtensionModulesState.OK
        else:
            return ExtensionModulesState.INVALID_VERSION
    return ExtensionModulesState.UNAVAILABLE


def _get_current_time() -> str:
    return "UTC " + str(datetime.utcnow())


def _get_installation_type() -> InstallationType:
    if os.path.exists("/.dockerenv"):
        return InstallationType.DOCKER
    else:
        return InstallationType.PYPI


def _get_installed_python_modules() -> List[str]:
    installed_packages = []
    for name, module in sorted(sys.modules.items()):
        if hasattr(module, "__version__"):
            installed_packages.append(f"{name}=={module.__version__}")  # pyre-ignore
        else:
            installed_packages.append(name)
    return installed_packages


def _get_startup_param() -> str:
    return " ".join(sys.argv[1:])


def _load_status_from_file(use_ext: bool, src_ver: str, ip_address: str, continent: str, country: str,
                           update_required: bool, account_id: Optional[str], quota_level: Optional[int]) -> Diagnostics:
    path = config.get_data_file(STATUS_FILE_NAME)
    with open(path, "r", encoding="utf-8") as json_file:
        status_file = json_file.read()
    try:
        model_dict = json.loads(status_file)
        diagnostics = model_loader.load_model(Diagnostics, model_dict)
    except JSONDecodeError:
        logger.warning(log_messages.STATUS_FILE_JSON_LOAD_FAIL, path)
        diagnostics = initialize(use_ext, src_ver, ip_address, continent, country, update_required, account_id, quota_level)
    return diagnostics


def _save_status_to_file(diagnostics: Diagnostics) -> None:
    path = config.get_data_file(STATUS_FILE_NAME)
    with open(path, "w", encoding="utf-8") as outfile:
        json.dump(diagnostics, outfile, cls=EnhancedJSONEncoder, indent=2)



import functools
import threading
from collections import deque
from concurrent.futures import Future
from threading import Thread, Condition, RLock
from typing import List, Callable, Any, NamedTuple, Deque, Optional


from bxutils import logging

logger = logging.get_logger(__name__)


class WorkItem(NamedTuple):
    callback: Callable
    future: Future


def handle_work_item(work_item: Optional[WorkItem]) -> None:
    if work_item is not None and not work_item.future.cancelled():
        try:
            result = work_item.callback()
        except BaseException as e:
            work_item.future.set_exception(e)
        else:
            try:
                work_item.future.set_result(result)
            # pylint: disable=broad-except
            except Exception as e:
                logger.exception(
                    "{} - unhandled error: {}, failed to set task result {}",
                    threading.current_thread().name,
                    e,
                    result
                )


class ThreadPool:

    def __init__(self, workers: int, thread_name_prefix: str) -> None:
        self._workers = workers
        self._stop_requested = False
        self._threads: List[Thread] = []
        self._work_items: Deque[WorkItem] = deque()
        self._condition = Condition(RLock())
        self._started = False
        for idx in range(self._workers):
            thread = Thread(name="{}-{}".format(thread_name_prefix, idx), target=self._thread_loop)
            thread.daemon = True
            self._threads.append(thread)

    def __enter__(self):
        if not self._started:
            self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def start(self) -> None:
        for thread in self._threads:
            thread.start()
        self._started = True

    def stop(self) -> None:
        with self._condition:
            self._stop_requested = True
            self._condition.notify_all()

    def close(self) -> None:
        if not self._stop_requested:
            self.stop()
        if self._started:
            for thread in self._threads:
                thread.join()
        while self._work_items:
            handle_work_item(self._work_items.popleft())

    def submit(self, callback: Callable[..., Any], *args, **kwargs) -> Future:
        if not self._started:
            raise RuntimeError("Thread pool was never started!")
        future = Future()
        work_item = WorkItem(functools.partial(callback, *args, **kwargs), future)
        with self._condition:
            self._work_items.append(work_item)
            self._condition.notify(1)
        return future

    def _thread_loop(self) -> None:
        try:
            while not self._stop_requested:
                work_item = None
                with self._condition:
                    if len(self._work_items) == 0:
                        self._condition.wait()
                    if self._work_items:
                        work_item = self._work_items.popleft()
                handle_work_item(work_item)
        # pylint: disable=broad-except
        except Exception as e:
            logger.fatal(
                "{} - unhandled error: {}, raised during thread execution!",
                threading.current_thread().name,
                e
            )



ONT_HASH_LEN = 32

ONT_TX_DEPLOY_TYPE_INDICATOR = 0xD0
ONT_TX_DEPLOY_TYPE_INDICATOR_AS_BYTEARRAY = bytearray([ONT_TX_DEPLOY_TYPE_INDICATOR])
ONT_TX_INVOKE_TYPE_INDICATOR = 0xD1
ONT_TX_INVOKE_TYPE_INDICATOR_AS_BYTEARRAY = bytearray([ONT_TX_INVOKE_TYPE_INDICATOR])



#!/bin/env python
#
# Copyright (C) 2018, bloXroute Labs, All rights reserved.
# See the file COPYING for details.
#
# Startup script for Gateway nodes
#
import argparse
import os
import random
import sys
import functools
from typing import Optional, List

from bxcommon import node_runner, constants
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils import cli, convert, config, ip_resolver
from bxcommon.models.node_type import NodeType

from bxutils import logging_messages_utils

from bxgateway import btc_constants, gateway_constants, ont_constants
from bxgateway.connections.gateway_node_factory import get_gateway_node_type
from bxgateway.testing.test_modes import TestModes
from bxcommon.utils.blockchain_utils.eth import crypto_utils, eth_common_constants
from bxgateway.gateway_opts import GatewayOpts
from bxgateway import argument_parsers
from bxgateway.utils.gateway_start_args import GatewayStartArgs
from bxgateway import gateway_init_tasks

MAX_NUM_CONN = 8192
PID_FILE_NAME = "bxgateway.pid"


def convert_net_magic(magic):
    if magic in btc_constants.BTC_MAGIC_NUMBERS:
        return btc_constants.BTC_MAGIC_NUMBERS[magic]
    elif magic in ont_constants.ONT_MAGIC_NUMBERS:
        return ont_constants.ONT_MAGIC_NUMBERS[magic]
    else:
        return int(magic)


def generate_default_nonce():
    return random.randint(0, sys.maxsize)


def parse_peer_string(peer_string):
    """
    Parses string of format ip:port:node_type,ip:port,ip:port:node_type,... to list of OutboundPeerModels.
    """
    peers = []
    for ip_port_string in peer_string.split(","):
        if ip_port_string:
            ip, port_str, node_type_str = ip_port_string.strip().split(":")
            peers.append(OutboundPeerModel(ip, int(port_str), node_type=NodeType[node_type_str.upper()]))
    return peers


def parse_streaming_peer(peer_string: str) -> OutboundPeerModel:
    ip, port_str = peer_string.split(":")
    return OutboundPeerModel(ip, int(port_str), node_type=NodeType.EXTERNAL_GATEWAY)


def get_default_eth_private_key():
    gateway_key_file_name = config.get_data_file(eth_common_constants.GATEWAY_PRIVATE_KEY_FILE_NAME)

    if os.path.exists(gateway_key_file_name):
        with open(gateway_key_file_name, "r") as key_file:
            private_key = key_file.read().strip()
    else:
        private_key = crypto_utils.generate_random_private_key_hex_str()
        with open(gateway_key_file_name, "w") as key_file:
            key_file.write(private_key)

    return private_key


def get_argument_parser() -> argparse.ArgumentParser:
    # Parse gateway specific command line parameters
    arg_parser = argparse.ArgumentParser(parents=[cli.get_argument_parser()],
                                         description="Command line interface for the bloXroute Gateway.",
                                         usage="bloxroute_gateway --blockchain-protocol [PROTOCOL] [additional "
                                               "arguments]")
    cli.add_argument_parser_rpc(arg_parser)

    arg_parser.add_argument("--blockchain-protocol", help="Blockchain protocol. e.g BitcoinCash, Ethereum", type=str)
    arg_parser.add_argument("--blockchain-network", help="Blockchain network. e.g Mainnet, Testnet", type=str)
    arg_parser.add_argument("--blockchain-port", help="Blockchain node port", type=int)
    arg_parser.add_argument("--blockchain-ip", help="Blockchain node ip",
                            type=ip_resolver.blocking_resolve_ip,
                            default=None)
    arg_parser.add_argument("--peer-gateways",
                            help="Optional gateway peer ip/ports that will always be connected to. "
                                 "Should be in the format ip1:port1:GATEWAY,ip2:port2:GATEWAY",
                            type=parse_peer_string,
                            default="")
    arg_parser.add_argument("--min-peer-gateways",
                            help="Minimum number of peer gateways before node will contact SDN for more.",
                            type=int,
                            default=0)
    arg_parser.add_argument("--remote-blockchain-ip", help="Remote blockchain node ip to proxy messages from",
                            type=ip_resolver.blocking_resolve_ip)
    arg_parser.add_argument("--remote-blockchain-port", help="Remote blockchain node port to proxy messages from",
                            type=int)
    arg_parser.add_argument("--connect-to-remote-blockchain",
                            help="If gateway should proxy messages from a remote bloXroute owned blockchain node",
                            type=convert.str_to_bool,
                            default=True)
    arg_parser.add_argument("--encrypt-blocks",
                            help="If gateway should encrypt blocks",
                            type=convert.str_to_bool,
                            default=False)
    arg_parser.add_argument("--peer-relays",
                            help="(TEST ONLY) Optional relays peer ip/ports that will always be connected to. "
                                 "Should be in the format ip1:port1:RELAY,ip2:port2:RELAY,...",
                            type=parse_peer_string,
                            default="")
    arg_parser.add_argument("--test-mode",
                            help="(TEST ONLY) Test modes to run. Possible values: {0}".format(
                                [TestModes.DROPPING_TXS]
                            ),
                            default="",
                            nargs="*")

    # Bitcoin specific
    arg_parser.add_argument("--blockchain-version", help="Bitcoin protocol version", type=int)
    arg_parser.add_argument("--blockchain-nonce", help="Bitcoin nonce", default=generate_default_nonce())
    arg_parser.add_argument("--blockchain-net-magic", help="Bitcoin net.magic parameter",
                            type=convert_net_magic)
    arg_parser.add_argument("--blockchain-services", help="Bitcoin services parameter", type=int)
    arg_parser.add_argument("--enable-node-cache", help="Retrieve peers from cookie if unavailable",
                            type=convert.str_to_bool,
                            default=True)

    # Ethereum specific
    arg_parser.add_argument("--node-public-key", help="Public key of Ethereum node for encrypted communication",
                            type=str)
    arg_parser.add_argument("--private-key", help="Private key for encrypted communication with Ethereum node",
                            type=str)
    arg_parser.add_argument("--network-id", help="Ethereum network id", type=int)
    arg_parser.add_argument("--genesis-hash", help="Genesis block hash of Ethereum network", type=str)
    arg_parser.add_argument("--chain-difficulty", help="Difficulty of genesis block Ethereum network (hex)", type=str)
    arg_parser.add_argument("--no-discovery", help="Disable discovery of Ethereum node and wait for node to connect",
                            type=bool, default=False)
    arg_parser.add_argument("--remote-public-key",
                            help="Public key of remote bloXroute owned Ethereum node for encrypted communication "
                                 "during chainstate sync ",
                            type=str)
    arg_parser.add_argument("--enode", help="Ethereum enode. ex) enode://<eth node public key>@<eth node "
                                            "ip>:<port>?discport=0",
                            action=argument_parsers.ParseEnode,
                            type=str,
                            default=None)

    # Blockchain peers specified by --blockchain-peers and --blockchain-peers-file would be aggregated together
    arg_parser.add_argument("--blockchain-peers",
                            help="A comma separated list of node peer info. "
                                 "For Ethereum, the format is enode://<eth node public key>@<eth node ip>:<port>. "
                                 "For other blockchain protocols, the format is <ip>:<port>",
                            action=argument_parsers.ParseBlockchainPeers,
                            type=str,
                            default=None)
    arg_parser.add_argument("--blockchain-peers-file",
                            help="A new line separated list of node peer info. "
                                 "For Ethereum, the format is enode://<eth node public key>@<eth node ip>:<port>. "
                                 "For other blockchain protocols, the format is <ip>:<port>",
                            type=str,
                            default=None)
    # IPC
    arg_parser.add_argument("--ipc",
                            help="Boolean indicating if IPC should be enabled.",
                            type=convert.str_to_bool,
                            default=False)
    arg_parser.add_argument("--ipc-file",
                            help="IPC filename that represents a unix domain socket",
                            type=str,
                            default="bxgateway.ipc")
    arg_parser.add_argument("--should-restart-on-high-memory",
                            help="Should a gateway restart itself if memory exceeds 2GB",
                            type=convert.str_to_bool,
                            default=True)
    # Ontology specific
    # TODO: Remove test only arguments
    arg_parser.add_argument("--http-info-port", help="(TEST ONLY)Ontology http server port to view node information",
                            type=int, default=config.get_env_default(GatewayStartArgs.HTTP_INFO_PORT))
    arg_parser.add_argument("--consensus-port", help="Ontology consensus port", type=int,
                            default=config.get_env_default(GatewayStartArgs.CONSENSUS_PORT))
    arg_parser.add_argument("--relay", help="(TEST ONLY)Ontology relay state", type=convert.str_to_bool,
                            default=config.get_env_default(GatewayStartArgs.RELAY_STATE))
    arg_parser.add_argument("--is-consensus", help="(TEST ONLY)Ontology consensus node", type=convert.str_to_bool,
                            default=config.get_env_default(GatewayStartArgs.IS_CONSENSUS))

    arg_parser.add_argument(
        "--compact-block",
        help="Specify either the gateway supports compact block message or not",
        type=convert.str_to_bool,
        default=constants.ACCEPT_COMPACT_BLOCK
    )
    arg_parser.add_argument(
        "--compact-block-min-tx-count",
        help="Minimal number of short transactions in compact block to attempt decompression.",
        type=int, default=btc_constants.BTC_COMPACT_BLOCK_DECOMPRESS_MIN_TX_COUNT
    )
    arg_parser.add_argument(
        "--dump-short-id-mapping-compression",
        help="If true, the gateway will dump all short ids and txhashes compressed into a block",
        type=convert.str_to_bool,
        default=False,
    )
    arg_parser.add_argument(
        "--dump-short-id-mapping-compression-path",
        help="Folder to dump compressed short ids to",
        default="/app/bxgateway/debug/compressed-short-ids",
    )
    arg_parser.add_argument(
        "--tune-send-buffer-size",
        help="If true, then the gateway will increase the send buffer's size for the blockchain connection",
        default=True,
        type=convert.str_to_bool,
    )
    arg_parser.add_argument(
        "--max-block-interval-s",
        help="Maximum time gateway holds a block while waiting for confirmation of receipt from blockchain node",
        type=int,
        default=gateway_constants.MAX_INTERVAL_BETWEEN_BLOCKS_S,
    )
    arg_parser.add_argument(
        "--cookie-file-path",
        help="Cookie file path",
        type=str,
    )
    arg_parser.add_argument(
        "--blockchain-message-ttl",
        help="Duration to queue up messages for if blockchain node connection is broken",
        type=int,
        default=gateway_constants.DEFAULT_BLOCKCHAIN_MESSAGE_TTL_S,
    )
    arg_parser.add_argument(
        "--remote-blockchain-message-ttl",
        help="Duration to queue up messages for if remote blockchain node connection is broken",
        type=int,
        default=gateway_constants.DEFAULT_REMOTE_BLOCKCHAIN_MESSAGE_TTL_S,
    )
    arg_parser.add_argument(
        "--stay-alive-duration",
        help="Duration Gateway should stay alive for without an active blockchain or relay connection",
        type=int,
        default=gateway_constants.DEFAULT_STAY_ALIVE_DURATION_S,
    )
    arg_parser.add_argument(
        "--initial-liveliness-check",
        help="Duration Gateway should stay alive for without an initial blockchain or relay connection",
        type=int,
        default=gateway_constants.INITIAL_LIVELINESS_CHECK_S,
    )
    arg_parser.add_argument(
        "--config-update-interval",
        help="update the node configuration on cron, 0 to disable",
        type=int,
        default=gateway_constants.CONFIG_UPDATE_INTERVAL_S,
    )
    arg_parser.add_argument(
        "--require-blockchain-connection",
        help="Close gateway if connection with blockchain node can't be established "
             "when the flag is set to True",
        type=convert.str_to_bool,
        default=True,
    )
    default_tx_flag = config.get_env_default(
        GatewayStartArgs.DEFAULT_TX_FLAG
    )
    arg_parser.add_argument(
        "--default-tx-flag",
        help=f"transaction flag to use when distributing transactions to the Bdn network (default: {default_tx_flag})",
        type=TransactionFlag.from_string,
        choices=list(TransactionFlag),
        default=default_tx_flag,
    )
    arg_parser.add_argument(
        "--ws",
        help=f"Enable RPC websockets server (default: False)",
        type=convert.str_to_bool,
        default=False,
    )
    arg_parser.add_argument(
        "--ws-host",
        help=f"Websockets server listening host (default: {gateway_constants.WS_DEFAULT_HOST})",
        type=str,
        default=gateway_constants.WS_DEFAULT_HOST,
    )
    arg_parser.add_argument(
        "--ws-port",
        help=f"Websockets server listening port (default: {gateway_constants.WS_DEFAULT_PORT}",
        type=int,
        default=gateway_constants.WS_DEFAULT_PORT,
    )
    arg_parser.add_argument(
        "--eth-ws-uri",
        help="Ethereum websockets endpoint for syncing transaction content",
        type=str,
    )
    arg_parser.add_argument(
        "--process-node-txs-in-extension",
        help="If true, then the gateway will process transactions received from blockchain node using C++ extension",
        default=True,
        type=convert.str_to_bool,
    )
    # TODO temp arg, need to be removed
    arg_parser.add_argument(
        "--enable-eth-extensions",
        help="If true, run compression and decompression using C++ extensions code",
        default=True,
        type=convert.str_to_bool,
    )
    arg_parser.add_argument(
        "--request-recovery",
        help="If true, gateway will send transaction recovery request to relay when unable to decompress a block",
        default=True,
        type=convert.str_to_bool,
    )
    arg_parser.add_argument(
        "--enable-block-compression",
        help="If set, overrides value from the SDN, and might be changed if SDN sent an update",
        type=convert.str_to_bool,
        nargs='?',
    )
    arg_parser.add_argument(
        "--filter-txs-factor",
        help="Ethereum only. Sets the factor of the average gas price to filter transactions below. "
             "(i.e. 0 => send all transactions (average * 0), 1 => send all transactions with average or "
             "higher gas price, etc.)",
        type=float,
        default=0
    )
    arg_parser.add_argument(
        "--stream-to-peer-gateway",
        help="Peer gateway to send confirmed transactions to from blockchain nodes. Requires --eth-ws-subscriber to be set.",
        type=parse_streaming_peer,
        default=None,
    )

    return arg_parser


def get_opts(args: Optional[List[str]] = None) -> GatewayOpts:
    config.set_working_directory(os.path.dirname(__file__))

    opts = GatewayOpts.from_opts(cli.parse_arguments(get_argument_parser(), args))
    config.set_data_directory(opts.data_dir)
    if opts.private_key is None:
        opts.private_key = get_default_eth_private_key()

    return opts


def main() -> None:
    logger_names = node_runner.LOGGER_NAMES.copy()
    logger_names.append("bxgateway")
    logging_messages_utils.logger_names = set(logger_names)
    opts = get_opts()
    get_node_class = functools.partial(get_gateway_node_type, opts.blockchain_protocol)

    node_runner.run_node(
        config.get_data_file(PID_FILE_NAME),
        opts,
        get_node_class,
        NodeType.EXTERNAL_GATEWAY,
        logger_names=logger_names,
        node_init_tasks=gateway_init_tasks.init_tasks
    )


if __name__ == "__main__":
    main()
